#
# Copyright (C) 2025 sits developers.
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <https://www.gnu.org/licenses/>.
#

"""Image utilities."""

import matplotlib.pyplot as plt
from PIL import Image


def show_local_image(image_path: str) -> None:
    """Displays a local image using matplotlib.

    Args:
        image_path (str): The file path to the image to be displayed.

    Returns:
        None: Nothing.
    """
    # load and crop image
    img = Image.open(image_path)
    img = img.crop(img.getbbox())

    # define figure object
    plt.figure(figsize=(7, 5), dpi=300)

    # show image in the figure canvas
    plt.imshow(img)

    # configure layout
    plt.axis("off")
    plt.tight_layout()

    # show image!
    plt.show()



#
# Copyright (C) 2025 sits developers.
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <https://www.gnu.org/licenses/>.
#

"""DSL for mask operations."""

from typing import Any

from pysits.conversions.dsl.base import RExpression


#
# Base R expression class
#
class MaskExpression(RExpression):
    """Mask expression class."""

    #
    # Magic methods
    #
    def __eq__(self, other):
        return Eq(self, other)

    def __ne__(self, other):
        return Neq(self, other)

    def __lt__(self, other):
        return Lt(self, other)

    def __le__(self, other):
        return Le(self, other)

    def __gt__(self, other):
        return Gt(self, other)

    def __ge__(self, other):
        return Ge(self, other)

    def __and__(self, other):
        return And(self, other)

    def __or__(self, other):
        return Or(self, other)

    def __invert__(self):  # for ~expr → maps to ``NOT``
        return Not(self)

    def __mod__(self, other: Any) -> None:
        """Disabled modulo operator to prevent confusion with R's %in% operator.

        Args:
            other: The right-hand operand (unused).

        Raises:
            NotImplementedError: Always raised to indicate that %in_ should
                                 be used instead.
        """
        raise NotImplementedError("Use %in_ for 'in' operation")

    #
    # Hashing
    #
    def __hash__(self) -> int:
        """Hash the expression."""
        return hash(self.r_repr())

    #
    # Abstract methods
    #
    def r_repr(self) -> str:
        """Convert the expression to its R string representation.

        Returns:
            str: The R code representation of the expression.

        Raises:
            NotImplementedError: This is an abstract method that must be implemented
                                 by subclasses.
        """
        raise NotImplementedError()


#
# Helper function
#
def as_rexpr(x: int | float | str | MaskExpression) -> MaskExpression:
    """Convert a Python value to an R expression.

    Args:
        x: Value to convert, can be an integer, float, string or
            existing ``MaskExpression``.

    Returns:
        MaskExpression: The value wrapped in an appropriate
            ``MaskExpression`` subclass.

    Raises:
        TypeError: If the input type cannot be converted to an
            ``MaskExpression``s.
    """
    if isinstance(x, MaskExpression):
        return x

    elif isinstance(x, (int | float | str)):
        return Literal(x)

    raise TypeError(f"Cannot convert {type(x)} to MaskExpression")


#
# Variable (symbol)
#
class MaskVariable(MaskExpression):
    """Represents an R variable or symbol."""

    def __init__(self, name: str) -> None:
        self.name = name

    def r_repr(self) -> str:
        """Convert the variable to its R string representation.

        Returns:
            str: The variable name as it appears in R.
        """
        return self.name

    # Add mask %in_ [...] syntactic sugar
    def in_(self, values: list[int | float | str]) -> "In":
        """Create an R %in% operation.

        Args:
            values: List of values to check membership against.

        Returns:
            In: An expression representing `var %in% c(values)` in R.
        """
        return In(self, values)


#
# Literal value
#
class Literal(MaskExpression):
    """Represents a literal value in R code.

    Args:
        value (Union[int, float, str]): The Python value to be represented in R.
    """

    def __init__(self, value: int | float | str) -> None:
        self.value = value

    def r_repr(self) -> str:
        """Convert the literal to its R string representation.

        Returns:
            str: The R code representation of the literal value.
        """
        return repr(self.value)


#
# Binary operator
#
class BinaryOp(MaskExpression):
    """Base class for binary operators.

    Args:
        left (Union[MaskExpression, int, float, str]): Left operand.

        right (Union[MaskExpression, int, float, str]): Right operand.
    """

    op: str | None = None
    """Operator."""

    def __init__(
        self,
        left: MaskExpression | int | float | str,
        right: MaskExpression | int | float | str,
    ) -> None:
        self.left = as_rexpr(left)
        self.right = as_rexpr(right)

    def r_repr(self) -> str:
        """Convert the binary operation to its R string representation.

        Returns:
            str: The R code representation of the binary operation.
        """
        return f"({self.left.r_repr()} {self.op} {self.right.r_repr()})"


#
# Specific binary ops
#
class Eq(BinaryOp):
    """Represents equality comparison (==) in R."""

    op = "=="


class Neq(BinaryOp):
    """Represents inequality comparison (!=) in R."""

    op = "!="


class Lt(BinaryOp):
    """Represents less than comparison (<) in R."""

    op = "<"


class Le(BinaryOp):
    """Represents less than or equal comparison (<=) in R."""

    op = "<="


class Gt(BinaryOp):
    """Represents greater than comparison (>) in R."""

    op = ">"


class Ge(BinaryOp):
    """Represents greater than or equal comparison (>=) in R."""

    op = ">="


class And(BinaryOp):
    """Represents logical AND (&) in R."""

    op = "&"


class Or(BinaryOp):
    """Represents logical OR (|) in R."""

    op = "|"


#
# Unary operator (NOT)
#
class Not(MaskExpression):
    """Represents logical NOT (!) in R.

    Args:
        expr (Union[RExpr, int, float, str]): Expression to negate.
    """

    def __init__(self, expr: MaskExpression | int | float | str) -> None:
        self.expr = as_rexpr(expr)

    def r_repr(self) -> str:
        """Convert the NOT operation to its R string representation.

        Returns:
            str: The R code representation of the NOT operation.
        """
        return f"!({self.expr.r_repr()})"


#
# In operator: mask %in% c(...)
#
class In(MaskExpression):
    """Represents the %in% operator in R.

    Args:
        left (Union[RExpr, int, float, str]): Value to check.

        values (List[Union[int, float, str]]): List of values to check against.
    """

    def __init__(
        self, left: MaskExpression | int | float | str, values: list[int | float | str]
    ) -> None:
        self.left = as_rexpr(left)
        self.values = values

    def r_repr(self) -> str:
        """Convert the %in% operation to its R string representation.

        Returns:
            str: The R code representation of the %in% operation.
        """
        values_str = ", ".join(repr(v) for v in self.values)
        return f"{self.left.r_repr()} %in% c({values_str})"


#
# List of expressions
#
class MaskExpressionList(MaskExpression):
    """Represents a named list of R expressions.

    Args:
        **entries (Dict[str, RExpr]): Keyword arguments mapping names to expressions.
    """

    def __init__(self, **entries: dict[str, MaskExpression]) -> None:
        self.entries = entries

    def r_repr(self) -> str:
        """Convert the expression list to its R string representation.

        Returns:
            str: The R code representation of the named list.
        """
        entries_str = ",\n    ".join(
            f'"{key}" = {expr.r_repr()}' for key, expr in self.entries.items()
        )
        return f"list(\n    {entries_str}\n)"


#
# Mask variables
#
MaskValue = MaskVariable("mask")
"""Mask variable for reclassify operations."""



#
# Copyright (C) 2025 sits developers.
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <https://www.gnu.org/licenses/>.
#

"""Cube operations."""

import rpy2.robjects as ro

from pysits.backend.pkgs import r_pkg_sits
from pysits.conversions.decorators import (
    function_call,
    rpy2_fix_type,
    rpy2_fix_type_custom,
)
from pysits.conversions.dsl.mask import MaskExpressionList
from pysits.docs import attach_doc
from pysits.models.data.cube import SITSCubeModel
from pysits.models.data.frame import SITSFrame


#
# Reclassify-specific converters functions
#
def convert_reclassify_rules(obj: object) -> MaskExpressionList:
    """Convert reclassify rules to a propert ExpressionList."""

    if not isinstance(obj, dict):
        raise ValueError("Reclassify rules must be a dictionary.")

    return MaskExpressionList(**obj)


#
# Reclassify-specific converters config
#
reclassify_converters = {
    "rules": convert_reclassify_rules,
}


#
# Cube
#
@function_call(r_pkg_sits.sits_cube, SITSCubeModel)
@attach_doc("sits_cube")
def sits_cube(*args, **kwargs) -> SITSCubeModel:
    """Create cubes."""


@function_call(r_pkg_sits.sits_regularize, SITSCubeModel)
@attach_doc("sits_regularize")
def sits_regularize(*args, **kwargs) -> SITSCubeModel:
    """Build a regular data cube from an irregular one."""


@function_call(r_pkg_sits.sits_cube_copy, SITSCubeModel)
@attach_doc("sits_cube_copy")
def sits_cube_copy(*args, **kwargs) -> SITSCubeModel:
    """Copy cubes."""


@function_call(r_pkg_sits.sits_mosaic, SITSCubeModel)
@attach_doc("sits_mosaic")
def sits_mosaic(*args, **kwargs) -> SITSCubeModel:
    """Mosaic classified cubes."""


#
# Probability and metrics
#
@function_call(r_pkg_sits.sits_variance, SITSCubeModel)
@attach_doc("sits_variance")
def sits_variance(*args, **kwargs) -> SITSCubeModel:
    """Calculate the variance of a probability cube."""


@function_call(r_pkg_sits.sits_uncertainty, SITSCubeModel)
@attach_doc("sits_uncertainty")
def sits_uncertainty(*args, **kwargs) -> SITSCubeModel:
    """Estimate classification uncertainty based on probs cube."""


@function_call(r_pkg_sits.sits_clean, SITSCubeModel)
@attach_doc("sits_clean")
def sits_clean(*args, **kwargs) -> SITSCubeModel:
    """Cleans a classified map using a local window."""


#
# Combine predictions
#
@function_call(r_pkg_sits.sits_combine_predictions, SITSCubeModel)
@attach_doc("sits_combine_predictions")
def sits_combine_predictions(*args, **kwargs) -> SITSCubeModel:
    """Estimate ensemble prediction based on list of probs cubes."""


#
# Sampling
#
@function_call(r_pkg_sits.sits_uncertainty_sampling, SITSFrame)
@attach_doc("sits_uncertainty_sampling")
def sits_uncertainty_sampling(*args, **kwargs) -> SITSFrame:
    """Suggest samples for enhancing classification accuracy."""


@function_call(r_pkg_sits.sits_confidence_sampling, SITSFrame)
@attach_doc("sits_confidence_sampling")
def sits_confidence_sampling(*args, **kwargs) -> SITSFrame:
    """Suggest high confidence samples to increase the training set."""


#
# Colors
#
@function_call(r_pkg_sits.sits_colors_qgis, lambda x: None)
@attach_doc("sits_colors_qgis")
def sits_colors_qgis(*args, **kwargs) -> None:
    """Function to save color table as QML style for data cube."""


#
# Base cube
#
@function_call(r_pkg_sits.sits_add_base_cube, SITSCubeModel)
@attach_doc("sits_add_base_cube")
def sits_add_base_cube(*args, **kwargs) -> SITSCubeModel:
    """Add base maps to a time series data cube."""


#
# Reclassify
#
@rpy2_fix_type_custom(converters=reclassify_converters)
@rpy2_fix_type
@attach_doc("sits_reclassify")
def sits_reclassify(
    cube: SITSCubeModel, mask: SITSCubeModel, rules: MaskExpressionList, *args, **kwargs
) -> SITSCubeModel:
    """Reclassify a classified cube."""
    params = []

    # Process parameters manually
    for k, v in kwargs.items():
        current_v = v[0]

        if k in ["output_dir", "version"]:
            current_v = f"'{current_v}'"

        elif k == "progress":
            current_v = "TRUE" if current_v else "FALSE"

        params.append(f"{k}={current_v}")

    # Build the ``sits_apply`` command manually to support
    # high-level expression definition (using string)
    command = f"""
        sits_reclassify(
            cube = {cube.r_repr()},
            mask = {mask.r_repr()},
            rules = {rules.r_repr()},
            {", ".join(params)}
        )
    """

    # Run operation
    result = ro.r(command)

    # Return
    return SITSCubeModel(result)


#
# Texture
#
@rpy2_fix_type
@attach_doc("sits_texture")
def sits_texture(cube, **kwargs) -> SITSCubeModel:
    """Apply a set of texture measures on a data cube."""
    params = []

    # Process parameters manually
    for k, v in kwargs.items():
        current_v = v[0]

        if k == "output_dir":
            current_v = f"'{current_v}'"

        elif k == "progress":
            current_v = "TRUE" if current_v else "FALSE"

        params.append(f"{k}={current_v}")

    # Build the ``sits_texture`` command manually to support
    # high-level expression definition (using string)
    command = f"""
        sits_texture(
            {cube.r_repr()},
            {", ".join(params)}
        )
    """

    # Run operation
    result = ro.r(command)

    # Return
    return SITSCubeModel(result)



#
# Copyright (C) 2025 sits developers.
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <https://www.gnu.org/licenses/>.
#

"""Xarray conversions."""

from __future__ import annotations

import dask.array as dask_array
import numpy as np
import rioxarray as xrio
import xarray as xr
from affine import Affine
from pandas import concat as pandas_concat
from pandas import to_datetime as pandas_to_datetime

from pysits.models.data.base import SITSData


#
# Auxiliary functions
#
def _xarray_load_raster(
    raster_path: str, crs: str, shape: tuple[int, int], transform: float
) -> xr.Dataset:
    """Load raster with rio-xarray.

    Args:
        raster_path (str): Complete path to the raster.

        crs (str): Dataset CRS.

        shape (Tuple[int, int]): Dataset shape.

        transform (float): Project transform.

    Returns:
        xr.Dataset: Raster loaded as Dataset
    """
    da = xrio.open_rasterio(raster_path, masked=True)
    da = da.squeeze("band", drop=True)

    return da.rio.reproject(
        dst_crs=crs,
        shape=shape,
        transform=transform,
        resampling=0,  # 0 = Nearest
    )


#
# SITS conversions function
#
def pandas_sits_as_xarray(data: SITSData) -> xr.Dataset:
    """Convert sits to xarray.

    Args:
        data (pysits.models.SITSData): SITS Data.

    Returns:
        xr.Dataset: SITS data as xarray.Dataset.
    """
    # Metadata columns
    time_series_metadata = data.drop(columns="time_series")

    # Extract time-series column
    time_series_data = data["time_series"]

    # Convert to a list of data frames
    time_series_data = time_series_data.tolist()

    # Get time-series attributes (removing ``Index``)
    time_series_attributes = set(time_series_data[0].columns).difference(["Index"])
    time_series_attributes = list(time_series_attributes)

    # Extract samples timeline
    timeline = time_series_data[0]["Index"]

    # Drop ``Index`` and create a stack
    time_series_data = np.stack(
        [ts.drop(columns="Index").to_numpy() for ts in time_series_data]
    )

    # Create xarray dataset
    return xr.Dataset(
        data_vars={
            var: (["sample", "time"], time_series_data[:, :, i])
            for i, var in enumerate(time_series_attributes)
        },
        coords={
            "sample": np.arange(len(time_series_metadata)),
            "time": timeline,
            "longitude": ("sample", time_series_metadata["longitude"].to_numpy()),
            "latitude": ("sample", time_series_metadata["latitude"].to_numpy()),
            "label": ("sample", time_series_metadata["label"].to_numpy()),
            "cube": ("sample", time_series_metadata["cube"].to_numpy()),
        },
    )


def pandas_cube_as_xarray(cube: SITSData) -> xr.Dataset:
    """Convert cube to xarray.

    Args:
        cube (pysits.models.SITSData): Cube data

    Returns:
        xr.Dataset: Cube data as xarray.Dataset.
    """
    # Get all files from the cube
    cube_file_info = cube["file_info"].tolist()

    # Merge and sort values
    cube_file_info = pandas_concat(cube_file_info, ignore_index=True)
    cube_file_info = cube_file_info.sort_values(["date", "band"]).reset_index(drop=True)

    # Assuming all cube have the same CRS / resolution, use one file
    # to extract ``shape``, ``coords`` and ``crs``
    cube_sample = cube_file_info.iloc[0]["path"]

    # Open file
    cube_sample = xrio.open_rasterio(cube_sample, masked=True)

    # Extract info
    cube_crs = cube_sample.rio.crs
    cube_res_x, cube_res_y = list(map(lambda x: abs(x), cube_sample.rio.resolution()))

    # To handle multiple tiles, use a ``global`` extent, covering all tiles
    xmin = cube_file_info["xmin"].min()
    xmax = cube_file_info["xmax"].max()
    ymin = cube_file_info["ymin"].min()
    ymax = cube_file_info["ymax"].max()

    # Calculate global shape
    width = int(np.ceil((xmax - xmin) / cube_res_x))
    height = int(np.ceil((ymax - ymin) / cube_res_y))

    # Define global transform
    global_transform = Affine.translation(xmin, ymax) * Affine.scale(
        cube_res_x, -cube_res_y
    )

    # Define X and Y coordinates
    x_coords = np.arange(width) * cube_res_x + xmin + cube_res_x / 2
    y_coords = ymax - np.arange(height) * cube_res_y - cube_res_y / 2

    # Prepare variables by band and date
    dataset_vars = {}

    for band, band_data in cube_file_info.groupby("band"):
        data_arrays = []
        time_coords = []

        # Sort by date
        band_data_sorted = band_data.sort_values("date")

        # Iterate bands
        for _, row in band_data_sorted.iterrows():
            # Get info
            da_path = row["path"]
            da_date = row["date"]

            # Load raster (band / date)
            da_raster = _xarray_load_raster(
                raster_path=da_path,
                crs=cube_crs,
                shape=(height, width),
                transform=global_transform,
            )

            # Save data
            data_arrays.append(da_raster)
            time_coords.append(da_date)

        # Stack data
        stacked = dask_array.stack(data_arrays, axis=0)

        # Save as variable (all bands in ``time``, ``y`` and ``x``)
        dataset_vars[band] = (("time", "y", "x"), stacked)

    # Build data cube
    ds = xr.Dataset(
        data_vars=dataset_vars,
        coords={
            "time": pandas_to_datetime(time_coords),
            "y": y_coords,
            "x": x_coords,
        },
    )

    # Save CRS
    ds.rio.write_crs(cube_crs, inplace=True)

    return ds



#
# Copyright (C) 2025 sits developers.
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <https://www.gnu.org/licenses/>.
#

"""DSL for tuning operations."""

from typing import Any

from pysits.conversions.dsl.base import RExpression


class TuningExpression(RExpression):
    """Tuning expression class."""


class TuningSymbol(TuningExpression):
    """Represents an R symbol (variable or function name)."""

    def __init__(self, name: str) -> None:
        """Initializer."""
        self.name = name

    def r_repr(self) -> str:
        """Convert the symbol to its R representation."""
        return self.name


class TuningFunctionCall(TuningExpression):
    """Represents an R function call."""

    def __init__(self, func: str, *args: Any, **kwargs: Any) -> None:
        """Initialize an R function call.

        Args:
            func: The name of the R function to call.

            *args: Positional arguments for the function call.

            **kwargs: Keyword arguments for the function call.
        """
        self.func = func
        self.args = args
        self.kwargs = kwargs

    def _convert_arg(self, arg: Any) -> str:
        """Convert a Python argument to its R representation.

        Args:
            arg: The argument to convert. Can be an RExpr, dict, str, bool, None,
                 or any other type that can be converted to string.

        Returns:
            str: The R code string representation of the argument.
        """

        def _convert_dict(d: dict) -> str:
            items = [f"{key} = {self._convert_arg(val)}" for key, val in d.items()]
            return f"list({', '.join(items)})"

        def _convert_string(s: str) -> str:
            if "::" in s or s.isidentifier():
                return TuningSymbol(s).r_repr()
            return f'"{s}"'

        def _convert_list(lvalue: list) -> str:
            return f"c({', '.join(map(str, lvalue))})"

        # Type converters
        # ToDo: Merge with conversions.base types
        type_converters = {
            TuningExpression: lambda x: x.r_repr(),
            dict: _convert_dict,
            str: _convert_string,
            list: _convert_list,
            tuple: _convert_list,
            bool: lambda x: "TRUE" if x else "FALSE",
            type(None): lambda _: "NULL",
        }

        converter = type_converters.get(type(arg))
        if converter is not None:
            return converter(arg)

        return str(arg)

    def r_repr(self) -> str:
        """Convert the function call to its R representation.

        Returns:
            str: The R code string representation of the function call.
        """
        args = [self._convert_arg(arg) for arg in self.args]
        args.extend(
            f"{key} = {self._convert_arg(val)}" for key, val in self.kwargs.items()
        )
        return f"{self.func}({', '.join(args)})"


def hparam(func_name: str, *args: Any, **kwargs: Any) -> TuningFunctionCall:
    """Create an R function call expression.

    This function creates an R function call expression that can be used to represent
    R function calls in Python code. The resulting expression can be converted to
    R code using the ``r_repr()`` method.

    Args:
        func_name: The name of the R function to call.

        *args: Positional arguments for the function call.

        **kwargs: Keyword arguments for the function call.

    Returns:
        RCall: An R function call expression.
    """
    return TuningFunctionCall(func_name, *args, **kwargs)



#
# Copyright (C) 2025 sits developers.
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <https://www.gnu.org/licenses/>.
#

"""backend packages."""

from pysits.backend.loaders import load_package
from pysits.settings import __sitsver__

# system pakage
r_pkg_base = load_package("base")
r_pkg_grdevices = load_package("grDevices")

# sits package
r_pkg_sits = load_package("sits", min_version=__sitsver__)

# sits-dependencies packages
r_pkg_tibble = load_package("tibble")
r_pkg_leaflet = load_package("leaflet")
r_pkg_kohonen = load_package("kohonen")
r_pkg_sf = load_package("sf")
r_pkg_htmlwidgets = load_package("htmlwidgets")
r_pkg_arrow = load_package("arrow")



#
# Copyright (C) 2025 sits developers.
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <https://www.gnu.org/licenses/>.
#

"""Cube models."""

from pandas import DataFrame as PandasDataFrame
from pandas import Series as PandasSeries
from rpy2.robjects.vectors import DataFrame as RDataFrame

from pysits.backend.functions import r_fnc_set_column
from pysits.conversions.tibble_arrow import (
    pandas_cube_to_tibble_arrow,
    tibble_cube_to_pandas_arrow,
)
from pysits.models.data.frame import SITSFrame


#
# Base class
#
class SITSCubeItemModel(PandasSeries):
    """SITS Data Cube item model.

    Attributes:
        _instance (RDataFrame): R DataFrame instance containing the cube data.

        required_columns (list[str]): List of required column names for a valid
            cube item.
    """

    #
    # Attributes
    #
    _instance: RDataFrame = None
    """R DataFrame instance."""

    required_columns: list[str] = [
        "source",
        "collection",
        "satellite",
        "sensor",
        "tile",
        "xmin",
        "xmax",
        "ymin",
        "ymax",
        "crs",
        "file_info",
    ]
    """Required columns for a valid cube item."""

    def __init__(self, data: PandasSeries, **kwargs) -> None:
        """Initialize a SITS Data Cube item.

        Args:
            data (PandasSeries): Input data containing cube metadata and information.

            **kwargs: Additional keyword arguments passed to pandas.Series constructor.

        Note:
            The input data must contain all required columns specified in
            ``required_columns``. If the data is valid, it will be converted
            to an R DataFrame instance.
        """
        # Create a cube from a Pandas Series
        if isinstance(data, PandasSeries) and getattr(data, "_instance", None) is None:
            # Check if required columns are present
            has_required_columns = all(
                col in data.index for col in self.required_columns
            )

            if has_required_columns:
                # Convert to Pandas DataFrame
                cube_data = PandasDataFrame([data])

                # Convert to R DataFrame
                self._instance = pandas_cube_to_tibble_arrow(cube_data)

        # Initialize super class
        super().__init__(data=data, **kwargs)


class SITSCubeModel(SITSFrame):
    """SITS Data Cube data.

    Attributes:
        _instance (RDataFrame): R DataFrame instance containing the cube data.
    """

    #
    # Attributes
    #
    required_columns: list[str] = [
        "source",
        "collection",
        "satellite",
        "sensor",
        "tile",
        "xmin",
        "xmax",
        "ymin",
        "ymax",
        "crs",
        "file_info",
    ]
    """Required columns for a valid cube."""

    #
    # Properties
    #
    @property
    def _constructor_sliced(self):
        """Return the constructor for sliced data."""
        return SITSCubeItemModel

    #
    # Dunder methods
    #
    def __init__(self, instance, **kwargs):
        """Initializer."""
        # If instance is a Pandas DataFrame, convert to R cube
        if isinstance(instance, PandasDataFrame):
            # Check if required columns are present
            has_required_columns = all(
                col in instance.columns for col in self.required_columns
            )

            if has_required_columns:
                self._instance = pandas_cube_to_tibble_arrow(instance)

        else:
            self._instance = instance

        # Proxy instance
        if isinstance(instance, RDataFrame):
            instance = self._convert_from_r(instance)

        # Initialize super class
        PandasDataFrame.__init__(self, data=instance, **kwargs)

    #
    # Convertions
    #
    def _convert_from_r(self, instance: RDataFrame) -> PandasDataFrame:
        """Convert data from R to Python.

        Args:
            instance (rpy2.robjects.vectors.DataFrame): Data instance.
        """
        return tibble_cube_to_pandas_arrow(instance)

    #
    # Data management
    #
    def _sync_instance(self):
        """Sync instance with R."""
        if not self._is_updated:
            return

        # Save current classes
        classes = self._instance.rclass

        # Update instance
        base_info = None

        if "base_info" in self.columns:
            # Convert each dataframe in the series to R DataFrame
            base_info = [pandas_cube_to_tibble_arrow(df) for df in self.base_info]

            # Drop base_info
            self.drop(columns=["base_info"], inplace=True)

        self._instance = pandas_cube_to_tibble_arrow(self)

        # Add base_info
        if base_info is not None:
            self._instance = r_fnc_set_column(self._instance, "base_info", base_info)

        # Restore classes
        self._instance.rclass = classes

    #
    # Representation
    #
    def _repr_html_(self) -> str:
        """Create an HTML representation of the cube."""
        from pysits.jinja import get_template

        # Get the HTML for the cube table from pandas
        super_html = super()._repr_html_() if hasattr(super(), "_repr_html_") else ""

        # Render the template
        return get_template("cube.html").render(
            cube_obj=self,
            super_html=super_html,
        )



#
# Copyright (C) 2025 sits developers.
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <https://www.gnu.org/licenses/>.
#

"""Unit tests for xarray export operations."""

import xarray as xr

from pysits.sits.context import samples_l8_rondonia_2bands
from pysits.sits.cube import sits_cube
from pysits.sits.exporters.xarray import sits_as_xarray
from pysits.sits.utils import r_package_dir


def test_xarray_cube_conversion():
    """Test xarray cube conversion."""
    # Create a data cube from local files
    data_dir = r_package_dir("extdata/raster/mod13q1", package="sits")
    cube = sits_cube(
        source="BDC",
        collection="MOD13Q1-6.1",
        data_dir=data_dir,
    )

    xcube = sits_as_xarray(cube)

    assert isinstance(xcube, xr.Dataset)
    assert xcube.time.size == 12  # noqa: PLR2004
    assert "NDVI" in xcube


def test_xarray_sits_conversion():
    """Test xarray sits conversion."""
    xsamples_ts = sits_as_xarray(samples_l8_rondonia_2bands)

    assert "EVI" in xsamples_ts
    assert "NDVI" in xsamples_ts
    assert xsamples_ts.time.size == 25  # noqa: PLR2004
    assert isinstance(xsamples_ts, xr.Dataset)



#
# Copyright (C) 2025 sits developers.
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <https://www.gnu.org/licenses/>.
#

"""Base DSL classes."""


class DSLObject:
    """Base class for DSL classes."""


class RExpression(DSLObject):
    def r_repr(self) -> str:
        """Convert the expression to its R representation.

        Returns:
            str: The R code string representation of the expression.

        Raises:
            NotImplementedError: This method must be implemented by subclasses.
        """
        raise NotImplementedError()

    def __str__(self) -> str:
        """String representation of the R expression.

        Returns:
            str: The R code string representation.
        """
        return self.r_repr()



#
# Copyright (C) 2025 sits developers.
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <https://www.gnu.org/licenses/>.
#

"""Common conversions."""

from datetime import date, timedelta
from pathlib import Path, PosixPath

import rpy2.robjects as ro
from geopandas import GeoDataFrame as GeoPandasDataFrame
from pandas import DataFrame as PandasDataFrame
from rpy2.rinterface_lib.sexp import NULLType
from rpy2.robjects import pandas2ri
from rpy2.robjects.robject import RObjectMixin

from pysits.backend.pkgs import r_pkg_tibble
from pysits.conversions.dsl.base import DSLObject
from pysits.conversions.tibble import geopandas_to_tibble, pandas_to_tibble

#
# Type mapping dictionary
#
TYPE_CONVERSIONS = {
    list: lambda obj: convert_list_like_to_r(obj),
    tuple: lambda obj: convert_list_like_to_r(obj),
    set: lambda obj: convert_list_like_to_r(list(obj)),
    dict: lambda obj: convert_dict_like_to_r(obj),
    bool: lambda obj: ro.BoolVector([obj]),
    int: lambda obj: ro.IntVector([obj]),
    float: lambda obj: ro.FloatVector([obj]),
    str: lambda obj: ro.StrVector([obj]),
    Path: lambda obj: ro.StrVector([obj.as_posix()]),
    PosixPath: lambda obj: ro.StrVector([obj.as_posix()]),
    PandasDataFrame: lambda obj: r_pkg_tibble.as_tibble(pandas_to_tibble(obj)),
    GeoPandasDataFrame: lambda obj: geopandas_to_tibble(obj),
}


#
# R epoch reference
#
EPOCH_START = date(1970, 1, 1)


#
# Base utilities
#
def _is_numeric(x: int | float) -> bool:
    """Helper function to check if a value is numeric (int or float)."""
    return isinstance(x, int | float)


#
# Base conversions
#
def convert_list_like_to_r(obj: list | tuple | set) -> ro.vectors.Vector:  # noqa: PLR0911
    """
    Converts a list-like object to the appropriate R vector type.

    Args:
        obj (list, tuple, set): The list-like Python object.

    Returns:
        An R-compatible vector.
    """
    if not obj:
        return ro.vectors.ListVector({})

    if all(isinstance(x, int) for x in obj):
        return ro.vectors.IntVector(obj)

    elif all(isinstance(x, float) for x in obj):
        return ro.vectors.FloatVector(obj)

    # Check if all elements are numeric (int or float)
    elif all(_is_numeric(x) for x in obj):
        return ro.vectors.FloatVector(obj)

    elif all(isinstance(x, str) for x in obj):
        return ro.vectors.StrVector(obj)

    elif all(isinstance(x, bool) for x in obj):
        return ro.BoolVector(obj)

    else:
        return ro.vectors.ListVector(
            {str(i): convert_to_r(v) for i, v in enumerate(obj)}
        )


def convert_dict_like_as_list_to_r(obj: dict) -> ro.vectors.ListVector:
    """Convert a Python dict to a ListVector."""
    return ro.vectors.ListVector({str(k): convert_to_r(v) for k, v in obj.items()})


def convert_dict_like_to_r(obj: dict) -> ro.vectors.Vector:
    """Convert a Python dictionary to an R vector.

    This function converts a Python dictionary to either a typed R vector or ListVector,
    depending on the types of values in the dictionary:
    - If all values are of the same type (e.g. all strings, all numeric), returns an
      appropriate typed R vector with named elements
    - Otherwise, returns an R ListVector with converted values

    Args:
        obj (dict): A Python dictionary to convert to an R vector.

    Returns:
        ro.vectors.Vector: Either a typed R vector (if all values are of same type) or
            an R ListVector (for mixed value types). The resulting vector will
            preserve the dictionary's keys as names in the R vector.
    """
    values = list(obj.values())
    if not values:
        return ro.vectors.ListVector({})

    # Get data properties
    has_unique_type = len(set(type(v) for v in values)) == 1
    is_numeric = all(_is_numeric(v) for v in values)

    # Handle homogeneous data
    if has_unique_type or is_numeric:
        vec = convert_list_like_to_r(values)
        vec.names = list(obj.keys())

        return vec

    # For mixed data, use a list vector
    return convert_dict_like_as_list_to_r(obj)


def convert_to_r(obj):
    """Convert Python objects to R-compatible objects for use with rpy2.

    Args:
        obj: The Python object to convert.

    Returns:
        An R-compatible object.

    Raises:
        TypeError: If the object type cannot be converted.
    """
    if obj is None:
        return ro.r("NULL")  # Convert None to R's NULL

    obj_type = type(obj)

    # Handle ``SITSBase`` objects
    if getattr(obj, "_instance", None):
        # Sync instance with R
        if getattr(obj, "_sync_instance", None):
            obj._sync_instance()

        return obj._instance

    # Handle ``raw R`` / Expressions objects
    if isinstance(obj, RObjectMixin | DSLObject):
        return obj

    # Check if the object type exists in the conversion dictionary
    if obj_type in TYPE_CONVERSIONS:
        return TYPE_CONVERSIONS[obj_type](obj)

    raise TypeError(f"Cannot convert object of type {obj_type} to R format")


def convert_to_python(obj, as_type="str"):
    """Convert an R object to a Python representation.

    Args:
        obj (ro.Vector): The R object to convert.

        as_type (str): The target Python type. Supported values:
            - ``str``: Converts R string vector to Python list of strings;

            - ``date``: Converts R DateVector to list of YYYY-MM-DD formatted strings;

            - ``int``: Converts R numeric vector to Python list of integers;

            - ``float``: Converts R numeric vector to Python list of floats;

            - ``bool``: Converts R logical vector to Python list of booleans.

    Returns:
        list: Converted values as a Python list.

    Raises:
        ValueError: If the specified ``as_type`` is not supported.
    """

    def _convert(value, type_):
        result = []

        if isinstance(value, ro.ListVector):
            for k, v in value.items():
                result.append(
                    {
                        str(k): convert_to_python(v, type_),
                    }
                )

        elif as_type and isinstance(value, ro.Vector):
            for el in value:
                if as_type == "str":
                    result.append(str(el))

                elif as_type == "date":
                    result.append(
                        (EPOCH_START + timedelta(days=int(el))).strftime("%Y-%m-%d")
                    )

                elif as_type == "int":
                    result.append(int(el))

                elif as_type == "float":
                    result.append(float(el))

                elif as_type == "bool":
                    result.append(bool(el))

        elif isinstance(value, ro.Vector):
            result = pandas2ri.rpy2py(value)
            names = value.names

            if not isinstance(names, NULLType):
                result = {str(k): v for k, v in zip(names, result)}

        return result

    return _convert(obj, as_type)


#
# Parameters conversions
#
def fix_reserved_words_parameters(**kwargs) -> dict:
    """Fix reserved words parameters.

    Args:
        **kwargs: The keyword arguments to fix.

    Returns:
        dict: The fixed keyword arguments.
    """
    new_values = {}
    keys_to_remove = []

    for key, _ in kwargs.items():
        # Assuming all reserved words end with "_"
        if key.endswith("_"):
            # Save key to remove
            keys_to_remove.append(key)

            # Save new value
            new_values[key[:-1]] = kwargs[key]

    # Remove keys that were converted
    for key in keys_to_remove:
        kwargs.pop(key)

    # Add remaining kwargs
    if new_values:
        kwargs.update(new_values)

    return kwargs



#
# Copyright (C) 2025 sits developers.
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <https://www.gnu.org/licenses/>.
#

"""Pytest configuration file."""

import webbrowser

import matplotlib

#
# Set the backend to avoid plot windows
#
matplotlib.use("Agg")

#
# Import after setting visualization backend
#
import matplotlib.pyplot as plt
import pytest


@pytest.fixture
def no_plot_window(monkeypatch):
    """Fixture to prevent matplotlib plot windows from showing during tests."""
    monkeypatch.setattr(plt, "show", lambda: None)
    yield
    plt.close("all")


@pytest.fixture
def no_browser(monkeypatch):
    """Fixture to prevent webbrowser from opening during tests."""
    monkeypatch.setattr(webbrowser, "open", lambda x: None)
    monkeypatch.setattr(webbrowser, "open_new", lambda x: None)
    monkeypatch.setattr(webbrowser, "open_new_tab", lambda x: None)
    yield



#
# Copyright (C) 2025 sits developers.
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <https://www.gnu.org/licenses/>.
#

"""Leaflet utilities."""

import os
import webbrowser

from rpy2.robjects.robject import RObject
from rpy2.robjects.vectors import ListVector

from pysits.backend.pkgs import r_pkg_htmlwidgets, r_pkg_sits

# Optional: only import IPython if available
try:
    from IPython.display import IFrame, display

    _IPYTHON_AVAILABLE = True

except ImportError:
    _IPYTHON_AVAILABLE = False


def _show_leaflet_widget(
    r_leaflet_widget: ListVector, width: str = "100%", height: str = "600px"
) -> None:
    """Save and display an R htmlwidget (leaflet map).

    Args:
        r_leaflet_widget: R object of class ``leaflet``/``htmlwidget`` (e.g. from
                          ``sits_view``).

        width: Width for display in Jupyter notebook (default: ``'100%'``).

        height: Height for display in Jupyter notebook (default: ``'600px'``)

    Returns:
        None: Nothing.
    """
    # Save to temporary HTML file
    tmp_path = "sits-map.html"

    # Save using R's saveWidget
    r_pkg_htmlwidgets.saveWidget(r_leaflet_widget, tmp_path, selfcontained=True)

    # Detect if we're in Jupyter (simple heuristic)
    in_jupyter = False
    try:
        from IPython import get_ipython

        if "IPKernelApp" in get_ipython().config:
            in_jupyter = True

    except Exception:  # noqa: E722
        pass

    # Display accordingly
    if in_jupyter and _IPYTHON_AVAILABLE:
        display(IFrame(src=tmp_path, width=width, height=height))

    else:  # noqa: E722
        webbrowser.open("file://" + os.path.realpath(tmp_path))

    # Remove the temporary file
    # os.unlink(tmp_path)


def plot_leaflet(instance: RObject, **kwargs) -> None:
    """Plot a leaflet map.

    Args:
        instance: The instance to plot.

        **kwargs: Additional keyword arguments passed to ``sits_view``.
    """
    # Generate the R plot
    leaflet_plot = r_pkg_sits.sits_view(instance, **kwargs)

    # Display the plot
    _show_leaflet_widget(leaflet_plot)



#
# Copyright (C) 2025 sits developers.
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <https://www.gnu.org/licenses/>.
#

"""tibble conversions."""

import warnings
from collections.abc import Callable

from geopandas import GeoDataFrame as GeoPandasDataFrame
from pandas import DataFrame as PandasDataFrame
from pandas import to_datetime as pandas_to_datetime
from rpy2 import robjects
from rpy2.robjects import StrVector, pandas2ri
from rpy2.robjects.conversion import localconverter
from rpy2.robjects.vectors import DataFrame as RDataFrame
from shapely import wkt

from pysits.backend.functions import r_fnc_class
from pysits.backend.pkgs import r_pkg_base, r_pkg_sf
from pysits.models.frame import SITSFrameArray


#
# Auxiliary functions
#
def _column_to_datetime(data: PandasDataFrame, colname: str) -> PandasDataFrame:
    """Transform a columns from R to a valid datetime column in Python.

    Args:
        data (pandas.DataFrame): Pandas Data Frame from an R tibble/data.frame.

        colname (str): Column to be converted datetime (from R format to Python format).

    Returns:
        pandas.DataFrame: Pandas data frame with ``colname`` as datetime.
    """
    # Convert if column is available
    if colname in data.columns:
        data[colname] = pandas_to_datetime(data[colname], origin="1970-01-01", unit="D")

    # Return!
    return data


def _sf_to_shapely(sf_object: RDataFrame) -> list:
    """Transform a columns from R to a valid geometry column in Python.

    Args:
        sf_object (rpy2.robjects.vectors.ListVector): R (tibble/data.frame) Data frame.

    Returns:
        list: List of Shapely geometries.
    """
    # Extract geometry and convert to WKT in R directly
    geom_wkt = r_pkg_sf.st_as_text(r_pkg_sf.st_geometry(sf_object))

    # Convert R character vector to Python list of strings
    geom_wkt_py = list(geom_wkt)

    # Convert each WKT string to a Shapely geometry
    return [wkt.loads(g) for g in geom_wkt_py]


#
# Base conversion function
#
def _tibble_to_pandas(
    data: RDataFrame,
    nested_columns: list | None = None,
    table_processor: Callable[[PandasDataFrame], PandasDataFrame] | None = None,
    nested_processor: Callable[[PandasDataFrame], PandasDataFrame] | None = None,
) -> PandasDataFrame:
    """Convert an R tibble containing nested data frames to a Pandas DataFrame.

    Args:
        data (RDataFrame): An R tibble/data.frame object that contains
                           nested data frames.

        nested_columns (list): List of column names that contain nested
                               data frames.

        table_processor (Callable | None, optional):
            A function to process the main table after conversion. The function should
            take a pandas DataFrame as input and return a processed pandas DataFrame.
            Defaults to None.

        nested_processor (Callable | None, optional):
            A function to process each nested data frame after conversion. The function
            should take a pandas DataFrame as input and return a processed pandas
            DataFrame. Defaults to None.

    Returns:
        PandasDataFrame: A pandas DataFrame where the nested columns are converted to
                         SITSFrameArray objects containing the processed nested
                         data frames.
    """
    # Check if the data is an SF object
    has_geometries = "sf" in r_fnc_class(data)

    # Define shapely geometries and CRS
    shapely_crs = None
    shapely_geometries = None

    # Handle SF objects
    if has_geometries:
        # Convert geometry to Shapely geometries
        shapely_geometries = _sf_to_shapely(data)

        # Get CRS
        shapely_crs = r_pkg_sf.st_crs(data)

        # Check if CRS is available
        if "NULL" not in r_fnc_class(shapely_crs):
            shapely_crs = shapely_crs.rx2("wkt")[0]

        # Drop geometry column
        data = r_pkg_sf.st_drop_geometry(data)

    # Convert columns definitions
    nested_columns = nested_columns if nested_columns else []

    # Extract columns from the data
    data_columns = r_pkg_base.colnames(data)

    # Remove invalid columns
    data_columns_valid = []

    for data_column in data_columns:
        col = data.rx2(data_column)

        # Remove invalid columns
        if r_fnc_class(col[0])[0] not in ["function", "NULL"]:
            data_columns_valid.append(data_column)

    # Replace old data columns with the filtered one
    data_columns = data_columns_valid

    # If user define nested columns, verify if they are available.
    if nested_columns:
        # Filter available nested columns
        nested_columns = [v for v in nested_columns if v in data_columns]

        # If selected columns are available, remove them from the data columns
        # This allows us to handle them individually.
        if nested_columns:
            data_columns = list(set(data_columns).difference(nested_columns))

    # Select regular columns (using ``[]``) and convert to Pandas
    rdf_data = data.rx(StrVector(data_columns))
    rdf_data = pandas2ri.rpy2py(rdf_data)

    # Handle nested columns if available
    for nested_column in nested_columns:
        # Select nested column (using ``[[]]``)
        nested_column_data = data.rx2(nested_column)

        # As it is a nested column, handle it as a list of ``tibble/data.frame``
        nested_column_processed = []

        for nested_row in nested_column_data:
            # Convert to pandas
            nested_row_df = pandas2ri.rpy2py(nested_row)

            # If a processor function is available, apply data to use
            if nested_processor and isinstance(nested_row_df, PandasDataFrame):
                nested_row_df = nested_processor(nested_row_df)

            # Save
            nested_column_processed.append(nested_row_df)

        # Convert column to SITS Array and save to the main data frame
        rdf_data[nested_column] = SITSFrameArray(nested_column_processed)

    # If a processor function is available to the main table, use it
    if table_processor:
        rdf_data = table_processor(rdf_data)

    # Transform dataframe to geodataframe
    if shapely_geometries:
        rdf_data = GeoPandasDataFrame(
            rdf_data, geometry=shapely_geometries, crs=shapely_crs
        )

    # Return!
    return rdf_data


#
# General function
#
def tibble_to_pandas(data: RDataFrame) -> PandasDataFrame:
    """Convert any tibble to Pandas DataFrame.

    Args:
        data (rpy2.robjects.vectors.DataFrame): R (tibble/data.frame) Data frame.

    Returns:
        pandas.DataFrame: R Data Frame as Pandas.
    """

    # Define table processor
    def _table_processor(x):
        """Table processor."""
        # Update date columns
        x = _column_to_datetime(x, "start_date")
        x = _column_to_datetime(x, "end_date")

        return x

    # Convert and return
    return _tibble_to_pandas(
        data=data,
        table_processor=_table_processor,
    )


def tibble_nested_to_pandas(
    data: RDataFrame,
    nested_columns: list,
    table_processor: Callable[[PandasDataFrame], PandasDataFrame] | None = None,
    nested_processor: Callable[[PandasDataFrame], PandasDataFrame] | None = None,
) -> PandasDataFrame:
    """(Public) Convert an R tibble containing nested data frames to a Pandas DataFrame.

    Args:
        data (RDataFrame): An R tibble/data.frame object that contains
                           nested data frames.

        nested_columns (list): List of column names that contain nested
                               data frames.

        table_processor (Callable | None, optional):
            A function to process the main table after conversion. The function should
            take a pandas DataFrame as input and return a processed pandas DataFrame.
            Defaults to None.

        nested_processor (Callable | None, optional):
            A function to process each nested data frame after conversion. The function
            should take a pandas DataFrame as input and return a processed pandas
            DataFrame. Defaults to None.

    Returns:
        PandasDataFrame: A pandas DataFrame where the nested columns are converted to
                         SITSFrameArray objects containing the processed nested
                         data frames.
    """
    return _tibble_to_pandas(
        data=data,
        nested_columns=nested_columns,
        table_processor=table_processor,
        nested_processor=nested_processor,
    )


#
# SITS conversions function
#
def tibble_sits_to_pandas(data: RDataFrame) -> PandasDataFrame:
    """Convert sits tibble to Pandas Data Frame.

    Args:
        data (rpy2.robjects.vectors.DataFrame): R (tibble/data.frame) Data frame.

    Returns:
        pandas.DataFrame: R Data Frame as Pandas.
    """
    # Define column order (from R)
    column_order = [
        "longitude",
        "latitude",
        "start_date",
        "end_date",
        "label",
        "cube",
        "time_series",
    ]

    # Define nested columns
    nested_columns = ["time_series", "predicted"]

    # Define table processor
    def _table_processor(x):
        """Table processor."""
        # Update date columns
        x = _column_to_datetime(x, "start_date")
        x = _column_to_datetime(x, "end_date")

        x_columns = list(filter(lambda y: y in x.columns, column_order))
        x_columns = x_columns + list(set(x.columns).difference(x_columns))

        return x[x_columns]

    # Define nested processor
    def _nested_processor(x):
        """Nested processor."""
        return _column_to_datetime(x, "Index")

    # Convert and return
    return _tibble_to_pandas(
        data=data,
        nested_columns=nested_columns,
        table_processor=_table_processor,
        nested_processor=_nested_processor,
    )


#
# Cube conversions function
#
def tibble_cube_to_pandas(data: RDataFrame) -> PandasDataFrame:
    """Convert sits tibble to Pandas Data Frame.

    Args:
        data (rpy2.robjects.vectors.DataFrame): R (tibble/data.frame) Data frame.

    Returns:
        pandas.DataFrame: R Data Frame as Pandas.
    """
    # Define column order (from R)
    column_order = [
        "source",
        "collection",
        "satellite",
        "sensor",
        "tile",
        "xmin",
        "xmax",
        "ymin",
        "ymax",
        "crs",
        "labels",
        "file_info",
        "vector_info",
    ]

    # Define nested columns
    nested_columns = ["labels", "file_info", "vector_info"]

    # Define table processor
    def _table_processor(x):
        """Table processor."""
        columns_available = [v for v in column_order if v in x.columns]

        return x[columns_available]

    # Define nested processor
    def _nested_processor(x):
        """Nested processor."""
        x = _column_to_datetime(x, "date")
        x = _column_to_datetime(x, "start_date")
        x = _column_to_datetime(x, "end_date")

        return x

    # Convert and return
    return _tibble_to_pandas(
        data=data,
        nested_columns=nested_columns,
        table_processor=_table_processor,
        nested_processor=_nested_processor,
    )


#
# Pandas to R conversions
#
def pandas_to_tibble(data: PandasDataFrame) -> RDataFrame:
    """Convert a pandas DataFrame to an R DataFrame object.

    This function converts a pandas DataFrame to an R DataFrame using
    rpy2's conversion infrastructure. It handles the conversion context
    to ensure proper type mapping between Python and R objects.

    Args:
        data (pandas.DataFrame): The pandas DataFrame to convert to R.

    Returns:
        rpy2.robjects.vectors.DataFrame: The converted R DataFrame object.

    Notes:
        - The function uses rpy2's localconverter to ensure proper conversion context
        - Handles both DataFrame and non-DataFrame inputs
        - Preserves column names and data types where possible
        - For non-DataFrame inputs, falls back to rpy2's default converter
    """
    with localconverter(robjects.default_converter + pandas2ri.converter):
        return robjects.conversion.py2rpy(data)


def geopandas_to_tibble(data: GeoPandasDataFrame) -> RDataFrame:
    """Convert pandas DataFrame or GeoDataFrame to R DataFrame or sf object.

    Removes columns that contain embedded pandas DataFrames.
    """
    data = GeoPandasDataFrame(data)

    if data.crs is None:
        raise ValueError("GeoDataFrame must have a CRS")

    # Identify columns where no cell is a DataFrame
    safe_columns = [col for col in data.columns if not data[col].dtype.name == "sits"]

    # Warn if columns are dropped
    dropped_columns = set(data.columns) - set(safe_columns)
    if dropped_columns:
        warnings.warn(
            f"Warning: Dropping columns with embedded DataFrames: {dropped_columns}"
        )

    # Keep only safe columns
    data_safe = data[safe_columns].copy()

    # If GeoDataFrame, convert geometry to WKT and include geometry column
    if isinstance(data, GeoPandasDataFrame):
        geom_col = data.geometry.name
        data_safe[geom_col] = data.geometry.to_wkt()

    # Convert to R DataFrame
    with localconverter(robjects.default_converter + pandas2ri.converter):
        r_df = robjects.conversion.py2rpy(data_safe)

    # If GeoDataFrame, convert to sf
    if isinstance(data, GeoPandasDataFrame):
        r_df = r_pkg_sf.st_as_sf(
            r_df,
            wkt=robjects.StrVector([geom_col]),
            crs=robjects.StrVector([data.crs.to_wkt()]),
        )

    return r_df



#
# Copyright (C) 2025 sits developers.
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <https://www.gnu.org/licenses/>.
#

"""Unit tests for geopandas operations."""

from geopandas import GeoDataFrame as GeoPandasDataFrame

from pysits.models.data.frame import SITSFrameSF
from pysits.models.data.ts import SITSTimeSeriesSFModel
from pysits.sits.context import samples_l8_rondonia_2bands
from pysits.sits.cube import sits_cube
from pysits.sits.exporters import sits_as_geopandas
from pysits.sits.utils import r_package_dir


def test_cube_geopandas_export():
    """Test cube as geopandas."""
    # Load cube
    data_dir = r_package_dir(
        "extdata/raster/mod13q1",
        package="sits",
    )
    cube = sits_cube(
        source="BDC",
        collection="MOD13Q1-6.1",
        data_dir=data_dir,
    )

    # Export data
    cube_gdf = sits_as_geopandas(cube)

    # Check properties
    assert cube_gdf.crs is not None
    assert cube_gdf.geometry.name == "geometry"
    assert isinstance(cube_gdf, SITSFrameSF)
    assert isinstance(cube_gdf, GeoPandasDataFrame)


def test_ts_geopandas_export():
    """Test time-series data as geopandas."""
    # Export data
    samples_gdf = sits_as_geopandas(samples_l8_rondonia_2bands)

    # Check properties
    assert samples_gdf.crs is not None
    assert samples_gdf.geometry.name == "geometry"
    assert isinstance(samples_gdf, SITSTimeSeriesSFModel)
    assert isinstance(samples_gdf, GeoPandasDataFrame)



#
# Copyright (C) 2025 sits developers.
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <https://www.gnu.org/licenses/>.
#

"""Unit tests for cube operations."""

from pathlib import Path

from pandas import DataFrame as PandasDataFrame

from pysits.conversions.dsl.mask import MaskValue
from pysits.models.data.cube import SITSCubeModel
from pysits.models.data.frame import SITSFrame
from pysits.sits.cube import sits_cube, sits_reclassify
from pysits.sits.data import sits_bands, sits_bbox, sits_labels, sits_timeline
from pysits.sits.utils import r_package_dir


def test_sits_cube_data_structure():
    """Test data structure of sits_cube."""
    # Define a region of interest for the city of Sinop
    roi_sinop = {
        "lon_min": -56.87417,
        "lon_max": -54.63718,
        "lat_min": -12.17083,
        "lat_max": -11.02292,
    }

    # Call sits_cube
    cube = sits_cube(
        source="BDC",
        collection="MOD13Q1-6.1",
        bands=["NDVI", "EVI"],
        roi=roi_sinop,
        start_date="2013-09-14",
        end_date="2014-08-29",
        progress=False,
    )

    # Verify the result is a SITSCubeModel
    assert isinstance(cube, SITSCubeModel)

    # Check columns
    assert all(
        col in cube.columns
        for col in [
            "source",
            "collection",
            "satellite",
            "sensor",
            "tile",
            "xmin",
            "xmax",
            "ymin",
            "ymax",
            "crs",
            "file_info",
        ]
    )

    # Check bands
    cube_bands = sits_bands(cube)
    assert all(band in cube_bands for band in ["NDVI", "EVI"])

    # Check timeline
    cube_timeline = sits_timeline(cube)
    assert len(cube_timeline) == 23  # noqa: PLR2004


def test_sits_cube_bbox():
    """Test bbox of sits cube."""
    # Define a region of interest for the city of Sinop
    roi_sinop = {
        "lon_min": -56.87417,
        "lon_max": -54.63718,
        "lat_min": -12.17083,
        "lat_max": -11.02292,
    }

    # Create a cube
    cube = sits_cube(
        source="BDC",
        collection="MOD13Q1-6.1",
        bands=["NDVI", "EVI"],
        roi=roi_sinop,
        start_date="2013-09-14",
        end_date="2014-08-29",
        progress=False,
    )

    # Get bbox
    bbox = sits_bbox(cube)

    # Check bbox structure
    assert isinstance(bbox, SITSFrame)
    assert all(col in bbox.columns for col in ["xmin", "xmax", "ymin", "ymax"])

    # Check crs transformation
    bbox_4326 = sits_bbox(cube, as_crs="EPSG:4326")
    bbox_4326 = bbox_4326.iloc[0]

    assert round(bbox_4326["xmin"], 4) == -63.8507  # noqa: PLR2004
    assert round(bbox_4326["xmax"], 4) == -50.7713  # noqa: PLR2004
    assert round(bbox_4326["ymin"], 4) == -20.0  # noqa: PLR2004
    assert round(bbox_4326["ymax"], 4) == -10.0  # noqa: PLR2004
    assert bbox_4326["crs"] == "EPSG:4326"


def test_sits_cube_filter():
    """Test filtering of sits cube."""
    cube = sits_cube(
        source="AWS",
        collection="SENTINEL-2-L2A",
        tiles=("20LLP", "20LKP"),
        bands=("B02", "B8A", "B11", "CLOUD"),
        start_date="2018-06-30",
        end_date="2018-08-31",
    )

    cube_tile1 = cube.query('tile == "20LLP"')
    assert isinstance(cube_tile1, SITSCubeModel)
    assert cube_tile1.tile.iloc[0] == "20LLP"
    assert not isinstance(cube_tile1._instance, PandasDataFrame)

    cube_tile2 = cube.query('tile == "20LKP"')
    assert isinstance(cube_tile2, SITSCubeModel)
    assert cube_tile2.tile.iloc[0] == "20LKP"
    assert not isinstance(cube_tile2._instance, PandasDataFrame)


def s(tmp_path: Path):
    """Test reclassify of classified cube."""
    # Open mask map
    data_dir = r_package_dir("extdata/raster/prodes", package="sits")
    prodes2021 = sits_cube(
        source="USGS",
        collection="LANDSAT-C2L2-SR",
        data_dir=data_dir,
        parse_info=("X1", "X2", "tile", "start_date", "end_date", "band", "version"),
        bands="class",
        version="v20220606",
        labels={
            "1": "Forest",
            "2": "Water",
            "3": "NonForest",
            "4": "NonForest2",
            "6": "d2007",
            "7": "d2008",
            "8": "d2009",
            "9": "d2010",
            "10": "d2011",
            "11": "d2012",
            "12": "d2013",
            "13": "d2014",
            "14": "d2015",
            "15": "d2016",
            "16": "d2017",
            "17": "d2018",
            "18": "r2010",
            "19": "r2011",
            "20": "r2012",
            "21": "r2013",
            "22": "r2014",
            "23": "r2015",
            "24": "r2016",
            "25": "r2017",
            "26": "r2018",
            "27": "d2019",
            "28": "r2019",
            "29": "d2020",
            "31": "r2020",
            "32": "Clouds2021",
            "33": "d2021",
            "34": "r2021",
        },
        progress=False,
    )

    # Open classification map
    data_dir = r_package_dir("extdata/raster/classif", package="sits")
    ro_class = sits_cube(
        source="MPC",
        collection="SENTINEL-2-L2A",
        data_dir=data_dir,
        parse_info=("X1", "X2", "tile", "start_date", "end_date", "band", "version"),
        bands="class",
        labels={
            "1": "ClearCut_Fire",
            "2": "ClearCut_Soil",
            "3": "ClearCut_Veg",
            "4": "Forest",
        },
        progress=False,
    )

    # Reclassify
    ro_mask = sits_reclassify(
        cube=ro_class,
        mask=prodes2021,
        rules=dict(
            Old_Deforestation=MaskValue.in_(
                [
                    "d2007",
                    "d2008",
                    "d2009",
                    "d2010",
                    "d2011",
                    "d2012",
                    "d2013",
                    "d2014",
                    "d2015",
                    "d2016",
                    "d2017",
                    "d2018",
                    "r2010",
                    "r2011",
                    "r2012",
                    "r2013",
                    "r2014",
                    "r2015",
                    "r2016",
                    "r2017",
                    "r2018",
                    "d2019",
                    "r2019",
                    "d2020",
                    "r2020",
                    "r2021",
                ]
            ),
            Water_Mask=(MaskValue == "Water"),
            NonForest_Mask=MaskValue.in_(["NonForest", "NonForest2"]),
        ),
        memsize=4,
        multicores=2,
        output_dir=tmp_path,
        version="ex_reclassify",
    )

    assert isinstance(ro_mask, SITSCubeModel)
    assert ro_mask.shape[0] == 1  # noqa: PLR2004 - number of tiles
    assert len(sits_labels(ro_mask)) == 5  # noqa: PLR2004 - number of labels



#
# Copyright (C) 2025 sits developers.
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <https://www.gnu.org/licenses/>.
#

"""Time-series data models."""

from geopandas import GeoDataFrame as GeoPandasDataFrame
from pandas import DataFrame as PandasDataFrame
from pandas import Series as PandasSeries
from rpy2.robjects.vectors import DataFrame as RDataFrame

from pysits.conversions.tibble import tibble_sits_to_pandas
from pysits.conversions.tibble_arrow import (
    pandas_sits_to_tibble_arrow,
    tibble_sits_to_pandas_arrow,
)
from pysits.models.data.frame import SITSFrame, SITSFrameSF


#
# Time-series data class
#
class SITSTimeSeriesItemModel(PandasSeries):
    """SITS time-series item model."""

    #
    # Attributes
    #
    _instance: RDataFrame = None
    """R DataFrame instance."""

    required_columns: list[str] = [
        "start_date",
        "end_date",
        "time_series",
    ]
    """Required columns for a valid time-series item."""

    def __init__(self, data: PandasSeries, **kwargs):
        """Initializer."""
        # Create a cube from a Pandas Series
        if isinstance(data, PandasSeries):
            # Check if required columns are present
            has_required_columns = all(
                col in data.index for col in self.required_columns
            )

            if has_required_columns:
                # Convert to Pandas DataFrame
                ts_data = PandasDataFrame([data])

                # Convert to R DataFrame
                self._instance = pandas_sits_to_tibble_arrow(ts_data)

        # Initialize super class
        super().__init__(data=data, **kwargs)


class SITSTimeSeriesModel(SITSFrame):
    """Time-series base class."""

    def __init__(self, instance, **kwargs):
        """Initializer."""
        # If instance is a Pandas DataFrame, convert to R cube
        if isinstance(instance, PandasDataFrame):
            # Convert to R DataFrame
            self._instance = pandas_sits_to_tibble_arrow(instance)

        else:
            self._instance = instance

        # Proxy instance
        if isinstance(instance, RDataFrame):
            instance = self._convert_from_r(instance)

        # Initialize super class
        PandasDataFrame.__init__(self, data=instance, **kwargs)

    #
    # Properties
    #
    @property
    def _constructor_sliced(self):
        """Return the constructor for sliced data."""
        return SITSTimeSeriesItemModel

    #
    # Convertions
    #
    def _convert_from_r(self, instance: RDataFrame, **kwargs) -> PandasDataFrame:
        """Convert data from R to Python.

        Args:
            instance (rpy2.robjects.vectors.DataFrame): Data instance.
        """
        return tibble_sits_to_pandas_arrow(instance)

    #
    # Data management
    #
    def _sync_instance(self):
        """Sync instance with R."""
        if not self._is_updated:
            return

        # Save current classes
        classes = self._instance.rclass

        # Update instance
        self._instance = pandas_sits_to_tibble_arrow(self)

        # Restore classes
        self._instance.rclass = classes


class SITSTimeSeriesSFModel(SITSFrameSF):
    """SITS time-series model as sf."""

    def __init__(self, *args, **kwargs):
        """Initializer."""
        super().__init__(*args, **kwargs)

    #
    # Convertions
    #
    def _convert_from_r(self, instance: RDataFrame, **kwargs) -> GeoPandasDataFrame:
        """Convert data from R to Python.

        Args:
            instance (rpy2.robjects.vectors.DataFrame): Data instance.
        """
        return tibble_sits_to_pandas(instance)


class SITSTimeSeriesPatternsModel(SITSTimeSeriesModel):
    """SITS patterns model."""

    def __init__(self, *args, **kwargs):
        """Initializer."""
        super().__init__(*args, **kwargs)


class SITSTimeSeriesClassificationModel(SITSTimeSeriesModel):
    """SITS time-series classification model."""

    def __init__(self, *args, **kwargs):
        """Initializer."""
        super().__init__(*args, **kwargs)



#
# Copyright (C) 2025 sits developers.
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <https://www.gnu.org/licenses/>.
#

"""Xarray exporter."""

from functools import singledispatch

from pysits.models.data.cube import SITSCubeModel
from pysits.models.data.frame import SITSFrame
from pysits.models.data.ts import SITSTimeSeriesModel


@singledispatch
def _sits_as_xarray(data: SITSFrame):
    """sits as xarray dispatch."""
    raise NotImplementedError(
        f"There is no `sits_as_xarray` available for {type(data)}"
    )


@_sits_as_xarray.register
def _(data: SITSTimeSeriesModel):
    """Convert sits to xarray."""
    from pysits.conversions.xarray import pandas_sits_as_xarray

    return pandas_sits_as_xarray(data)


@_sits_as_xarray.register
def _(data: SITSCubeModel):
    """Convert cube to xarray."""
    from pysits.conversions.xarray import pandas_cube_as_xarray

    return pandas_cube_as_xarray(data)


def sits_as_xarray(data: SITSFrame):
    """Convert data to xarray."""
    try:
        from pysits.conversions.xarray import (
            pandas_cube_as_xarray,  # noqa
            pandas_sits_as_xarray,  # noqa
        )

    except ImportError as e:
        raise ImportError(
            "xarray dependencies not installed. To use this feature, please install "
            "them with `pip install pysits[xarray]`."
        ) from e

    return _sits_as_xarray(data)



#
# Copyright (C) 2025 sits developers.
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <https://www.gnu.org/licenses/>.
#

"""Matrix data models."""

from pandas import DataFrame as PandasDataFrame
from rpy2.rinterface_lib.sexp import NULLType
from rpy2.robjects.vectors import FloatMatrix

from pysits.backend.functions import r_fnc_class
from pysits.conversions.common import convert_to_python
from pysits.conversions.vector import matrix_to_pandas
from pysits.models.data.base import SITSData
from pysits.models.data.frame import SITSFrame
from pysits.models.data.table import SITSTable
from pysits.models.data.vector import SITSNamedVector


class SITSMatrix(SITSFrame):
    """Base class for sits matrix results."""

    #
    # Dunder methods
    #
    def __init__(self, instance, **kwargs):
        """Initializer."""
        self._instance = instance

        # Proxy instance
        if "matrix" in r_fnc_class(instance):
            instance = self._convert_from_r(instance)

        # Initialize super class
        PandasDataFrame.__init__(self, data=instance, **kwargs)

    #
    # Convertions
    #
    def _convert_from_r(self, instance, **kwargs):
        """Convert data from R to Python."""
        return matrix_to_pandas(instance)


class SITSConfusionMatrix(SITSData):
    """Base class for sits confusion matrix results."""

    #
    # Convertions
    #
    def _convert_from_r(self, instance):
        """Convert data from R to Python."""
        return instance

    #
    # Properties
    #
    @property
    def positive(self) -> str | None:
        """Positive property."""
        value = self._instance.rx2("positive")
        is_null = isinstance(value, NULLType)

        return str(value[0]) if not is_null else None

    @property
    def table(self) -> SITSTable:
        """Table property."""
        return SITSTable(self._instance.rx2("table"))

    @property
    def overall(self) -> SITSNamedVector:
        """Overall property."""
        return SITSNamedVector(self._instance.rx2("overall"))

    @property
    def by_class(self) -> SITSNamedVector | SITSMatrix:
        """By class property."""
        value = self._instance.rx2("byClass")
        is_matrix = isinstance(value, FloatMatrix)

        return SITSNamedVector(value) if not is_matrix else SITSMatrix(value)

    @property
    def mode(self) -> str | None:
        """Mode property."""
        value = self._instance.rx2("mode")
        is_null = isinstance(value, NULLType)

        return str(value[0]) if not is_null else None

    @property
    def dots(self) -> list:
        """Dots property."""
        return convert_to_python(self._instance.rx2("dots"))

    #
    # Dunder methods
    #
    def __str__(self):
        """String representation."""
        return str(self._instance)

    def __repr__(self):
        """Representation."""
        return str(self._instance)



#
# Copyright (C) 2025 sits developers.
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <https://www.gnu.org/licenses/>.
#

"""Unit tests for sits utils."""

from pathlib import Path

import pytest

from pysits.backend.loaders import load_function_from_package
from pysits.models.data.ts import SITSTimeSeriesModel
from pysits.sits.context import samples_modis_ndvi
from pysits.sits.utils import (
    r_package_dir,
    r_set_seed,
    read_rds,
)


def test_r_set_seed():
    """Test r_set_seed."""
    assert r_set_seed(42) is None


def test_read_rds(tmp_path: Path):
    """Test read RDS compatible with SITS."""
    r_fnc_save_rds = load_function_from_package("base::saveRDS")

    # Create a temporary file
    rds_file = tmp_path / "samples_deforestation.rds"

    # Save RDS
    r_fnc_save_rds(samples_modis_ndvi._instance, rds_file.as_posix())

    # Read RDS
    rds_content = read_rds(rds_file)

    # Save RDS
    assert isinstance(rds_content, SITSTimeSeriesModel)


def test_read_rds_file_not_found():
    """Test read RDS file not found."""
    with pytest.raises(FileNotFoundError):
        read_rds("non-existing-file.rds")


def test_r_package_dir():
    """Test get package dir from an existing R package."""
    assert r_package_dir("extdata/raster/mod13q1", package="sits")



#
# Copyright (C) 2025 sits developers.
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <https://www.gnu.org/licenses/>.
#

"""pysits module."""

from .conversions.dsl.mask import MaskValue
from .conversions.dsl.tuning import hparam
from .settings import __version__
from .sits.classification import sits_classify, sits_label_classification, sits_smooth
from .sits.colors import (
    sits_colors,
    sits_colors_reset,
    sits_colors_set,
    sits_colors_show,
)
from .sits.config import sits_config, sits_config_show, sits_config_user_file
from .sits.context import (
    cerrado_2classes,
    point_mt_6bands,
    samples_l8_rondonia_2bands,
    samples_modis_ndvi,
)
from .sits.cube import (
    sits_add_base_cube,
    sits_clean,
    sits_colors_qgis,
    sits_combine_predictions,
    sits_confidence_sampling,
    sits_cube,
    sits_cube_copy,
    sits_mosaic,
    sits_reclassify,
    sits_regularize,
    sits_texture,
    sits_uncertainty,
    sits_uncertainty_sampling,
    sits_variance,
)
from .sits.data import (
    sits_accuracy,
    sits_accuracy_summary,
    sits_apply,
    sits_bands,
    sits_bbox,
    sits_labels,
    sits_labels_summary,
    sits_list_collections,
    sits_merge,
    sits_mixture_model,
    sits_reduce,
    sits_select,
    sits_timeline,
)
from .sits.data import sits_summary as summary
from .sits.exporters import sits_as_geopandas, sits_as_xarray, sits_to_csv, sits_to_xlsx
from .sits.impute import impute_linear, sits_impute
from .sits.ml import (
    sits_formula_linear,
    sits_formula_logref,
    sits_kfold_validate,
    sits_lightgbm,
    sits_lighttae,
    sits_mlp,
    sits_model_export,
    sits_resnet,
    sits_rfor,
    sits_svm,
    sits_tae,
    sits_tempcnn,
    sits_train,
    sits_xgboost,
)
from .sits.segment import sits_segment, sits_slic
from .sits.tiles import sits_mgrs_to_roi, sits_roi_to_tiles, sits_tiles_to_roi
from .sits.ts import (
    sits_cluster_clean,
    sits_cluster_dendro,
    sits_cluster_frequency,
    sits_geo_dist,
    sits_get_class,
    sits_get_data,
    sits_get_probs,
    sits_patterns,
    sits_pred_features,
    sits_pred_normalize,
    sits_pred_references,
    sits_pred_sample,
    sits_predictors,
    sits_reduce_imbalance,
    sits_sample,
    sits_sampling_design,
    sits_sgolay,
    sits_show_prediction,
    sits_som_clean_samples,
    sits_som_evaluate_cluster,
    sits_som_map,
    sits_stats,
    sits_stratified_sampling,
    sits_validate,
    sits_whittaker,
)
from .sits.tuning import sits_tuning, sits_tuning_hparams
from .sits.utils import (
    load_samples,
    r_package_dir,
    r_set_seed,
    read_rds,
)
from .sits.visualization import sits_plot as plot
from .sits.visualization import sits_view

__all__ = (
    # Classification
    "sits_classify",
    "sits_smooth",
    "sits_label_classification",
    # Cube
    "sits_cube",
    "sits_clean",
    "sits_combine_predictions",
    "sits_add_base_cube",
    "sits_reclassify",
    "sits_regularize",
    "sits_cube_copy",
    "sits_colors_qgis",
    "sits_uncertainty",
    "sits_uncertainty_sampling",
    "sits_confidence_sampling",
    "sits_variance",
    "sits_texture",
    "sits_mosaic",
    # Colors
    "sits_colors",
    "sits_colors_reset",
    "sits_colors_set",
    "sits_colors_show",
    # Configuration
    "sits_config",
    "sits_config_show",
    "sits_config_user_file",
    # Data management
    "sits_bands",
    "sits_timeline",
    "sits_labels",
    "sits_list_collections",
    "sits_bbox",
    "sits_select",
    "sits_merge",
    "sits_mixture_model",
    "summary",
    "sits_labels_summary",
    "sits_accuracy",
    "sits_accuracy_summary",
    "sits_cluster_dendro",
    "sits_cluster_frequency",
    "sits_cluster_clean",
    "sits_reduce",
    # Machine Learning methods
    "sits_train",
    "sits_mlp",
    "sits_rfor",
    "sits_resnet",
    "sits_tempcnn",
    "sits_lighttae",
    "sits_svm",
    "sits_xgboost",
    "sits_lightgbm",
    "sits_tae",
    "sits_kfold_validate",
    "sits_model_export",
    "sits_formula_linear",
    "sits_formula_logref",
    # Impute
    "sits_impute",
    "impute_linear",
    # Time-series
    "sits_show_prediction",
    "sits_sgolay",
    "sits_whittaker",
    "sits_get_data",
    "sits_get_class",
    "sits_get_probs",
    "sits_stats",
    "sits_validate",
    "sits_predictors",
    "sits_pred_features",
    "sits_pred_normalize",
    "sits_pred_references",
    "sits_pred_sample",
    "sits_som_map",
    "sits_som_evaluate_cluster",
    "sits_som_clean_samples",
    "sits_geo_dist",
    "sits_patterns",
    "sits_sample",
    "sits_reduce_imbalance",
    "sits_sampling_design",
    "sits_stratified_sampling",
    # Tiles
    "sits_mgrs_to_roi",
    "sits_tiles_to_roi",
    "sits_roi_to_tiles",
    # Segments
    "sits_segment",
    "sits_slic",
    # Apply
    "sits_apply",
    # Exporters
    "sits_to_csv",
    "sits_as_xarray",
    "sits_as_geopandas",
    "sits_to_xlsx",
    # Tuning
    "sits_tuning_hparams",
    "sits_tuning",
    # Visualization
    "plot",
    "sits_view",
    # DSL Variables
    "MaskValue",
    "hparam",
    # Context data
    "samples_l8_rondonia_2bands",
    "samples_modis_ndvi",
    "point_mt_6bands",
    "cerrado_2classes",
    # Utils
    "read_rds",
    "r_package_dir",
    "r_set_seed",
    "load_samples",
    # Package settings
    "__version__",
)



#
# Copyright (C) 2025 sits developers.
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <https://www.gnu.org/licenses/>.
#

"""Clojure utilities."""

from collections.abc import Callable
from typing import Any, ParamSpec, TypeVar

from pysits.backend.pkgs import r_pkg_sits
from pysits.conversions.decorators import rpy2_fix_type, rpy2_fix_type_custom

#
# Generics
#
R = TypeVar("R")
P = ParamSpec("P")


#
# Factory function
#
def closure_factory(
    name: str, converters: dict[str, Callable[[Any], Any]] = {}
) -> Callable[P, R]:
    """Factory to create sits-based closure functions.

    This function creates a closure that wraps R package sits functions,
    making them callable from Python. It checks if the requested function
    exists in the R sits package before creating the wrapper.

    Args:
        name (str): Name of the sits-based function to be wrapped from the R package.

    Returns:
        Callable[P, R]: A wrapped function that forwards calls to the R sits package.
            The function accepts any arguments (*args, **kwargs) and returns the result
            from the R function.

    Raises:
        ValueError: If the specified function name does not exist in the R sits package.
    """
    if not hasattr(r_pkg_sits, name):
        raise ValueError(f"Invalid function: {name}")

    # define method closure
    @rpy2_fix_type_custom(converters)
    @rpy2_fix_type
    def _fnc(*args: P.args, **kwargs: P.kwargs) -> R:
        return getattr(r_pkg_sits, name)(*args, **kwargs)

    # set function name
    _fnc.__name__ = name

    return _fnc



#
# Copyright (C) 2025 sits developers.
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <https://www.gnu.org/licenses/>.
#

"""Unit tests for visualization operations."""

from pysits.sits.context import samples_l8_rondonia_2bands
from pysits.sits.cube import sits_cube
from pysits.sits.ml import sits_rfor, sits_train
from pysits.sits.ts import sits_patterns, sits_som_map
from pysits.sits.utils import r_package_dir
from pysits.sits.visualization import sits_plot, sits_view


def test_sits_visualization(no_plot_window):
    """Test sits visualization."""
    sits_plot(samples_l8_rondonia_2bands)


def test_sits_patterns_visualization(no_plot_window):
    """Test sits patterns visualization."""
    patterns = sits_patterns(samples_l8_rondonia_2bands)
    sits_plot(patterns)


def test_machine_learning_visualization(no_plot_window):
    """Test machine learning visualization."""
    ml_model = sits_train(samples_l8_rondonia_2bands, sits_rfor())
    sits_plot(ml_model)


def test_som_visualization(no_plot_window):
    """Test SOM visualization."""
    som = sits_som_map(data=samples_l8_rondonia_2bands)
    sits_plot(som)


def test_cube_visualization(no_plot_window):
    """Test cube visualization."""
    cube = sits_cube(
        source="BDC",
        collection="MOD13Q1-6.1",
        data_dir=r_package_dir("extdata/raster/mod13q1", package="sits"),
    )

    sits_plot(cube)


def test_classified_cube_visualization(no_plot_window):
    """Test classified cube visualization."""
    data_dir = r_package_dir("extdata/raster/classif", package="sits")
    cube = sits_cube(
        source="MPC",
        collection="SENTINEL-2-L2A",
        data_dir=data_dir,
        parse_info=(
            "X1",
            "X2",
            "tile",
            "start_date",
            "end_date",
            "band",
            "version",
        ),
        bands="class",
        labels={
            "1": "ClearCut_Fire",
            "2": "ClearCut_Soil",
            "3": "ClearCut_Veg",
            "4": "Forest",
        },
        progress=False,
    )

    # Plot the result
    sits_plot(cube)


def test_sits_visualization_leaflet(no_browser):
    """Test sits visualization."""
    sits_view(samples_l8_rondonia_2bands)


def test_cube_visualization_leaflet(no_browser):
    """Test cube visualization."""
    # Create a cube
    cube = sits_cube(
        source="BDC",
        collection="MOD13Q1-6.1",
        data_dir=r_package_dir("extdata/raster/mod13q1", package="sits"),
    )

    # Plot the cube
    sits_view(cube)



#
# Copyright (C) 2025 sits developers.
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <https://www.gnu.org/licenses/>.
#

"""Accuracy data models."""

from pandas import DataFrame as PandasDataFrame

from pysits.conversions.common import convert_to_python
from pysits.models.data.base import SITSData
from pysits.models.data.matrix import SITSConfusionMatrix, SITSMatrix
from pysits.models.data.table import SITSTable
from pysits.models.data.vector import SITSNamedVector


class SITSAccuracy(SITSData):
    """Base class for sits accuracy results."""

    #
    # Convertions
    #
    def _convert_from_r(self, instance):
        """Convert data from R to Python."""
        return instance

    #
    # Properties
    #
    @property
    def error_matrix(self) -> SITSConfusionMatrix:
        """Error matrix property."""
        return SITSTable(self._instance.rx2("error_matrix"))

    @property
    def area_pixels(self) -> SITSTable:
        """Area pixels property."""
        return SITSNamedVector(self._instance.rx2("area_pixels"))

    @property
    def error_ajusted_area(self) -> SITSNamedVector:
        """Error adjusted area property."""
        return SITSNamedVector(self._instance.rx2("error_ajusted_area"))

    @property
    def stderr_prop(self) -> SITSNamedVector | SITSMatrix:
        """Standard error property."""
        return SITSNamedVector(self._instance.rx2("stderr_prop"))

    @property
    def stderr_area(self) -> SITSNamedVector | SITSMatrix:
        """Standard error area property."""
        return SITSNamedVector(self._instance.rx2("stderr_area"))

    @property
    def conf_interval(self) -> SITSNamedVector | SITSMatrix:
        """Confidence interval property."""
        return SITSNamedVector(self._instance.rx2("conf_interval"))

    @property
    def accuracy(self) -> SITSNamedVector:
        """Accuracy property."""
        value = convert_to_python(self._instance.rx2("accuracy"), as_type=None)
        value = {k: v for d in value for k, v in d.items()}

        # Updating user and producer values
        value["user"] = PandasDataFrame(value["user"], index=[0])
        value["producer"] = PandasDataFrame(value["producer"], index=[0])

        # Return
        return value

    #
    # Dunder methods
    #
    def __str__(self):
        """String representation."""
        return str(self._instance)

    def __repr__(self):
        """Representation."""
        return str(self._instance)



#
# Copyright (C) 2025 sits developers.
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <https://www.gnu.org/licenses/>.
#

"""Unit tests for data operations (cube and time-series)."""

from pathlib import Path

from pysits.models.data.cube import SITSCubeModel
from pysits.models.data.ts import SITSTimeSeriesModel
from pysits.sits.context import samples_l8_rondonia_2bands
from pysits.sits.cube import sits_cube, sits_regularize
from pysits.sits.data import (
    sits_apply,
    sits_bands,
    sits_merge,
    sits_reduce,
    sits_select,
    sits_timeline,
)
from pysits.sits.utils import r_package_dir


def test_cube_select():
    """Test cube select."""

    cube = sits_cube(
        source="BDC",
        collection="CBERS-WFI-16D",
        bands=("NDVI", "EVI"),
        tiles="007004",
        start_date="2018-09-01",
        end_date="2019-08-28",
    )

    # Select bands
    cube_ndvi = sits_select(cube, bands="NDVI")
    assert isinstance(cube_ndvi, SITSCubeModel)
    assert sits_bands(cube_ndvi) == ["NDVI"]

    # Select time
    cube_ndvi_short = sits_select(
        cube_ndvi, start_date="2018-08-29", end_date="2019-01-01"
    )
    assert len(sits_timeline(cube_ndvi_short)) == 9  # noqa: PLR2004


def test_sits_select():
    """Test sits select."""

    # Select bands
    samples_evi = sits_select(samples_l8_rondonia_2bands, bands="EVI")
    assert isinstance(samples_evi, SITSTimeSeriesModel)
    assert sits_bands(samples_evi) == ["EVI"]

    # Select time
    samples_evi_short = sits_select(
        samples_evi, start_date="2019-01-17", end_date="2019-07-12"
    )
    assert len(sits_timeline(samples_evi_short)) == 12  # noqa: PLR2004


def test_merge_cubes(tmp_path: Path):
    """Test merging of Sentinel-1 and Sentinel-2 cubes."""
    dir1 = tmp_path / "s1"
    dir2 = tmp_path / "s2"
    dir1.mkdir(parents=True, exist_ok=True)
    dir2.mkdir(parents=True, exist_ok=True)

    # Create Sentinel-1 RTC cube
    cube_s1_rtc = sits_cube(
        source="MPC",
        collection="SENTINEL-1-RTC",
        bands=("VV", "VH"),
        orbit="descending",
        tiles=("22LBL"),
        start_date="2021-06-01",
        end_date="2021-10-01",
        progress=False,
    )

    # Create Sentinel-2 cube
    cube_s2 = sits_cube(
        source="MPC",
        collection="SENTINEL-2-L2A",
        bands=("B02", "B8A", "B11", "CLOUD"),
        tiles=("22LBL"),
        start_date="2021-06-01",
        end_date="2021-09-30",
        progress=False,
    )

    # Regularize Sentinel-1 cube
    cube_s1_reg = sits_regularize(
        cube=cube_s1_rtc,
        period="P16D",
        res=540,
        tiles=("22LBL"),
        memsize=4,
        multicores=4,
        output_dir=dir2,
        progress=False,
    )

    # Regularize Sentinel-2 cube
    cube_s2_reg = sits_regularize(
        cube=cube_s2,
        period="P16D",
        res=540,
        memsize=4,
        multicores=4,
        output_dir=dir1,
        progress=False,
    )

    # Merge the regularized cubes
    cube_s1_s2 = sits_merge(cube_s2_reg, cube_s1_reg)

    # Verify the result is a SITSCubeModel
    assert isinstance(cube_s1_s2, SITSCubeModel)

    # Check that the merged cube contains all bands
    merged_bands = sits_bands(cube_s1_s2)
    assert all(band in merged_bands for band in ["VV", "VH", "B02", "B8A", "B11"])

    # Check tiles
    assert cube_s1_s2["tile"].iloc[0] == "22LBL"

    # Check timeline
    timeline = sits_timeline(cube_s1_s2)
    assert len(timeline) > 0  # Should have at least one date
    assert min(timeline) >= "2021-06-01"
    assert max(timeline) <= "2021-09-22"


def test_sits_apply():
    """Test sits apply."""
    points = sits_select(samples_l8_rondonia_2bands, bands="NDVI")
    points_norm = sits_apply(
        points, NDVI_norm="(NDVI - min(NDVI)) / (max(NDVI) - min(NDVI))"
    )

    points_nonnorm = sits_apply(
        points,
        NDVI_nonnorm="(NDVI - min(NDVI)) / (max(NDVI) - min(NDVI))",
        normalized=False,
    )

    assert all(band in sits_bands(points_norm) for band in ["NDVI", "NDVI_norm"])
    assert all(band in sits_bands(points_nonnorm) for band in ["NDVI", "NDVI_nonnorm"])


def test_cube_apply(tmp_path: Path):
    """Test cube apply."""
    data_dir = r_package_dir(
        "extdata/raster/mod13q1",
        package="sits",
    )
    cube = sits_cube(
        source="BDC",
        collection="MOD13Q1-6.1",
        data_dir=data_dir,
    )

    # Generate a texture images with variance in NDVI images
    cube_texture = sits_apply(
        data=cube, NDVITEXTURE="w_median(NDVI)", window_size=5, output_dir=tmp_path
    )

    assert all(band in sits_bands(cube_texture) for band in ["NDVI", "NDVITEXTURE"])


def test_sits_reduce():
    """Test reduce in samples."""
    points = sits_select(samples_l8_rondonia_2bands, "NDVI")
    points_reduced = sits_reduce(points, NDVI_MEDIAN="t_median(NDVI)")

    assert "NDVI-MEDIAN" in sits_bands(points_reduced)
    assert len(sits_timeline(points_reduced)) == 1  # noqa: PLR2004 - one date


def test_cube_reduce(tmp_path: Path):
    """Test reduce in cube."""
    data_dir = r_package_dir(
        "extdata/raster/mod13q1",
        package="sits",
    )
    cube = sits_cube(
        source="BDC",
        collection="MOD13Q1-6.1",
        data_dir=data_dir,
    )

    # Reduce the cube
    cube_reduced = sits_reduce(
        cube,
        NDVIMEAN="t_mean(NDVI)",
        output_dir=tmp_path,
        progress=False,
        multicores=2,
    )

    # Check cube properties
    assert "NDVIMEAN" in sits_bands(cube_reduced)
    assert len(sits_timeline(cube_reduced)) == 1  # noqa: PLR2004 - one date



#
# Copyright (C) 2025 sits developers.
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <https://www.gnu.org/licenses/>.
#

"""backend functions."""

from pysits.backend.loaders import load_function_from_package

# Base - plot
r_fnc_plot = load_function_from_package("base::plot")

# Base - set column
r_fnc_set_column = load_function_from_package("base::$<-")

# Base - summary (base)
r_fnc_summary = load_function_from_package("base::summary")

# Base - readRDS (base)
r_fnc_read_rds = load_function_from_package("base::readRDS")

# Base - system.file (base)
r_fnc_system_file = load_function_from_package("base::system_file")

# Base - set.seed (base)
r_fnc_set_seed = load_function_from_package("base::set_seed")

# Base - class (base)
r_fnc_class = load_function_from_package("base::class")

# Base - as.data.frame (base)
r_fnc_as_data_frame = load_function_from_package("base::as_data_frame")

# Base - colnames (base)
r_fnc_colnames = load_function_from_package("base::colnames")

# Base - rownames (base)
r_fnc_rownames = load_function_from_package("base::rownames")



#
# Copyright (C) 2025 sits developers.
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <https://www.gnu.org/licenses/>.
#

"""Unit tests for data operations (cube and time-series)."""

import pytest


def test_lazy_import():
    """Test lazy import of extras."""
    from pysits.extras import earthdatalogin_edl_netrc

    assert earthdatalogin_edl_netrc is not None


def test_lazy_import_error():
    """Test lazy import of extras."""

    with pytest.raises(ImportError):
        from pysits.extras import invalidmodule_function_name  # noqa: F401


def test_earthdatalogin():
    """Test earthdatalogin functions."""
    from pysits.extras import earthdatalogin_edl_netrc

    assert earthdatalogin_edl_netrc(username="test", password="test")


def test_torch():
    """Test torch functions."""
    from pysits.extras import torch_install_torch

    assert torch_install_torch()



#
# Copyright (C) 2025 sits developers.
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <https://www.gnu.org/licenses/>.
#

"""Tile-related operations."""

from pysits.backend.pkgs import r_pkg_sits
from pysits.conversions.decorators import function_call
from pysits.docs import attach_doc
from pysits.models.data.frame import SITSFrameSF
from pysits.models.data.vector import SITSNamedVector


@function_call(r_pkg_sits.sits_mgrs_to_roi, SITSNamedVector)
@attach_doc("sits_mgrs_to_roi")
def sits_mgrs_to_roi(*args, **kwargs) -> SITSNamedVector:
    """Convert MGRS to ROI."""


@function_call(r_pkg_sits.sits_tiles_to_roi, SITSNamedVector)
@attach_doc("sits_tiles_to_roi")
def sits_tiles_to_roi(*args, **kwargs) -> SITSNamedVector:
    """Convert tiles to ROI."""


@function_call(r_pkg_sits.sits_roi_to_tiles, SITSFrameSF)
@attach_doc("sits_roi_to_tiles")
def sits_roi_to_tiles(*args, **kwargs) -> SITSFrameSF:
    """Find tiles of a given ROI and Grid System."""



#
# Copyright (C) 2025 sits developers.
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR ANY PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <https://www.gnu.org/licenses/>.
#

"""Frame data accessor for SITS (Satellite Image Time Series) data management.

This module provides a custom pandas DataFrame accessor that enables seamless
interaction with SITS data frames. The accessor provides methods to manage
sample properties such as labels, ensuring proper integration between Python
pandas DataFrames and the underlying R SITS package.

The SITSAccessor class extends pandas DataFrames with SITS-specific functionality,
allowing users to work with satellite image time series data in a familiar
pandas interface while maintaining compatibility with the R SITS ecosystem.
"""

import pandas as pd
from pandas.api.extensions import register_dataframe_accessor
from rpy2.robjects.vectors import StrVector

from pysits.backend.pkgs import r_pkg_sits
from pysits.conversions.common import convert_to_python


@register_dataframe_accessor("sits")
class SITSAccessor:
    """Custom pandas DataFrame accessor for SITS (Satellite Image Time Series) data.

    This accessor provides SITS-specific functionality to pandas DataFrames,
    enabling management of data properties such as sample labels.

    The accessor is automatically registered with pandas and can be accessed
    via the `.sits` attribute on any DataFrame that contains SITS data.

    Attributes:
        _obj (pd.DataFrame): The pandas DataFrame object that this accessor
            is attached to.

    Example:
        >>> import pysits
        >>> # Assuming df is a DataFrame with SITS data
        >>> df.sits.labels  # Get current labels
        >>> df.sits.labels = ["forest", "deforestation"]  # Set new labels
    """

    def __init__(self, pandas_obj: pd.DataFrame) -> None:
        """Initialize the SITS accessor with a pandas DataFrame.

        Args:
            pandas_obj: The pandas DataFrame to attach this accessor to.
        """
        self._obj = pandas_obj

    def _check_instance(self) -> bool:
        """Check if the DataFrame contains valid SITS data.

        This method verifies that the DataFrame has an `_instance` attribute
        that points to a valid R SITS object. This is a prerequisite for
        all SITS-specific operations.

        Returns:
            True if the DataFrame contains valid SITS data.

        Raises:
            ValueError: If the DataFrame does not contain valid SITS data
                (i.e., missing `_instance` attribute).
        """
        if getattr(self._obj, "_instance", None):
            return True

        raise ValueError("Data is not a SITS data frame.")

    @property
    def labels(self) -> list[str]:
        """Get the sample labels from the SITS data frame.

        This property retrieves the current sample labels from the underlying
        R SITS object and converts them to a Python list of strings.

        Returns:
            A list of strings representing the current sample labels in the
            SITS data frame.

        Raises:
            ValueError: If the DataFrame does not contain valid SITS data.

        Example:
            >>> df.sits.labels
            ['forest', 'deforestation', 'water']
        """
        self._check_instance()
        return convert_to_python(r_pkg_sits.sits_labels(self._obj._instance))

    @labels.setter
    def labels(self, new_labels: list[str] | tuple) -> None:
        """Set new sample labels for the SITS data frame.

        This setter updates the sample labels in the underlying R SITS object.

        Args:
            new_labels: A list or tuple of strings representing the new
                sample labels to assign to the SITS data frame.

        Raises:
            ValueError: If the DataFrame does not contain valid SITS data.

        Example:
            >>> df.sits.labels = ["forest", "deforestation", "water"]
            >>> df.sits.labels  # ['forest', 'deforestation', 'water']
        """
        self._check_instance()

        # Get label setter function from R SITS package
        r_set_labels_func = getattr(r_pkg_sits, "sits_labels<-")

        # Convert labels to R vector
        r_labels = StrVector(new_labels)

        # Set labels in the R SITS object
        self._obj._instance = r_set_labels_func(self._obj._instance, r_labels)



#
# Copyright (C) 2025 sits developers.
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <https://www.gnu.org/licenses/>.
#

"""Time-series operations."""

from pysits.backend.pkgs import r_pkg_sits
from pysits.conversions.common import convert_to_python
from pysits.conversions.decorators import function_call
from pysits.docs import attach_doc
from pysits.models.data.accuracy import SITSConfusionMatrix
from pysits.models.data.base import SITStructureData
from pysits.models.data.frame import SITSFrame, SITSFrameNested, SITSFrameSF
from pysits.models.data.matrix import SITSMatrix
from pysits.models.data.table import SITSTable
from pysits.models.data.ts import SITSTimeSeriesModel, SITSTimeSeriesPatternsModel


#
# High-level operation
#
@function_call(r_pkg_sits.sits_get_data, SITSTimeSeriesModel)
@attach_doc("sits_get_data")
def sits_get_data(*args, **kwargs) -> SITSTimeSeriesModel:
    """Retrieve time series data from a data cube."""


@function_call(r_pkg_sits.sits_get_class, SITSFrame)
@attach_doc("sits_get_class")
def sits_get_class(*args, **kwargs) -> SITSFrame:
    """Get values from classified maps."""


@function_call(r_pkg_sits.sits_get_probs, lambda x: SITSFrameNested(x, ["neighbors"]))
@attach_doc("sits_get_probs")
def sits_get_probs(*args, **kwargs) -> SITSFrameNested:
    """Get probabilities from classified maps."""


@function_call(r_pkg_sits.sits_stats, SITStructureData)
@attach_doc("sits_stats")
def sits_stats(*args, **kwargs) -> SITStructureData:
    """Obtain statistics for all sample bands.

    ToDo:
        - Enhance result type to a Dict-like object.
    """


#
# Validation
#
@function_call(r_pkg_sits.sits_validate, SITSConfusionMatrix)
@attach_doc("sits_validate")
def sits_validate(*args, **kwargs) -> SITSConfusionMatrix:
    """Validate time series samples."""


#
# Predictors
#
@function_call(r_pkg_sits.sits_predictors, SITSFrame)
@attach_doc("sits_predictors")
def sits_predictors(*args, **kwargs) -> SITSFrame:
    """Obtain predictors for time series samples."""


@function_call(r_pkg_sits.sits_pred_features, SITSFrame)
@attach_doc("sits_pred_features")
def sits_pred_features(*args, **kwargs) -> SITSFrame:
    """Obtain numerical values of predictors for time series samples."""


@function_call(r_pkg_sits.sits_pred_normalize, SITSFrame)
@attach_doc("sits_pred_normalize")
def sits_pred_normalize(*args, **kwargs) -> SITSFrame:
    """Normalize predictor values."""


@function_call(
    r_pkg_sits.sits_pred_references, lambda x: convert_to_python(x, as_type="str")
)
@attach_doc("sits_pred_references")
def sits_pred_references(*args, **kwargs) -> list[str]:
    """Obtain categorical id and predictor labels for time series samples."""


@function_call(r_pkg_sits.sits_pred_sample, SITSFrame)
@attach_doc("sits_pred_sample")
def sits_pred_sample(*args, **kwargs) -> SITSFrame:
    """Obtain a fraction of the predictors data frame."""


@function_call(r_pkg_sits.sits_sample, SITSTimeSeriesModel)
@attach_doc("sits_sample")
def sits_sample(*args, **kwargs) -> SITSTimeSeriesModel:
    """Sample a percentage of a time series."""


@function_call(r_pkg_sits.sits_reduce_imbalance, SITSTimeSeriesModel)
@attach_doc("sits_reduce_imbalance")
def sits_reduce_imbalance(*args, **kwargs) -> SITSTimeSeriesModel:
    """Reduce imbalance in a set of samples."""


@function_call(r_pkg_sits.sits_show_prediction, SITSFrame)
@attach_doc("sits_show_prediction")
def sits_show_prediction(*args, **kwargs) -> SITSFrame:
    """Show prediction results."""


#
# Sampling
#
@function_call(r_pkg_sits.sits_sampling_design, SITSMatrix)
@attach_doc("sits_sampling_design")
def sits_sampling_design(*args, **kwargs) -> SITSMatrix:
    """Allocation of sample size to strata."""


@function_call(r_pkg_sits.sits_stratified_sampling, SITSFrameSF)
@attach_doc("sits_stratified_sampling")
def sits_stratified_sampling(*args, **kwargs) -> SITSFrameSF:
    """Allocation of sample size to strata."""


#
# SOM
#
@function_call(r_pkg_sits.sits_som_map, SITStructureData)
@attach_doc("sits_som_map")
def sits_som_map(*args, **kwargs) -> SITStructureData:
    """Build a SOM for quality analysis of time series samples."""


@function_call(r_pkg_sits.sits_som_evaluate_cluster, SITSFrame)
@attach_doc("sits_som_evaluate_cluster")
def sits_som_evaluate_cluster(*args, **kwargs) -> SITSFrame:
    """Evaluate cluster quality."""


@function_call(r_pkg_sits.sits_som_clean_samples, SITSTimeSeriesModel)
@attach_doc("sits_som_clean_samples")
def sits_som_clean_samples(*args, **kwargs) -> SITSTimeSeriesModel:
    """Cleans the samples based on SOM map information."""


#
# Dendrogram
#
@function_call(r_pkg_sits.sits_cluster_dendro, SITSTimeSeriesModel)
@attach_doc("sits_cluster_dendro")
def sits_cluster_dendro(*args, **kwargs) -> SITSTimeSeriesModel:
    """Find clusters in time series samples.

    ToDo:
        - ToDo: Add support for the `plot` argument.
    """


@function_call(r_pkg_sits.sits_cluster_frequency, SITSTable)
@attach_doc("sits_cluster_frequency")
def sits_cluster_frequency(*args, **kwargs) -> SITSTable:
    """Show label frequency in each cluster produced by dendrogram analysis."""


@function_call(r_pkg_sits.sits_cluster_clean, SITSTimeSeriesModel)
@attach_doc("sits_cluster_clean")
def sits_cluster_clean(*args, **kwargs) -> SITSTimeSeriesModel:
    """Removes labels that are minority in each cluster."""


#
# Patterns
#
@function_call(r_pkg_sits.sits_patterns, SITSTimeSeriesPatternsModel)
@attach_doc("sits_patterns")
def sits_patterns(*args, **kwargs) -> SITSTimeSeriesPatternsModel:
    """Find temporal patterns associated to a set of time series."""


#
# Filtering
#
@function_call(r_pkg_sits.sits_sgolay, SITSTimeSeriesModel)
@attach_doc("sits_sgolay")
def sits_sgolay(*args, **kwargs) -> SITSTimeSeriesModel:
    """Apply Savitzky-Golay filter to time series."""


@function_call(r_pkg_sits.sits_whittaker, SITSTimeSeriesModel)
@attach_doc("sits_whittaker")
def sits_whittaker(*args, **kwargs) -> SITSTimeSeriesModel:
    """Apply Whittaker filter to time series."""


#
# Distances
#
@function_call(r_pkg_sits.sits_geo_dist, SITSFrame)
@attach_doc("sits_geo_dist")
def sits_geo_dist(*args, **kwargs) -> SITSFrame:
    """Compute the minimum distances among samples and prediction points."""



#
# Copyright (C) 2025 sits developers.
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <https://www.gnu.org/licenses/>.
#

"""Arrow conversions."""

import os
import tempfile
from collections.abc import Callable

from pandas import DataFrame as PandasDataFrame
from pandas.core.generic import NDFrame as PandasNDFrame
from pyarrow import feather
from rpy2.rinterface_lib.sexp import NULLType
from rpy2.robjects import StrVector, pandas2ri
from rpy2.robjects import globalenv as rpy2_globalenv
from rpy2.robjects import r as rpy2_r_interface
from rpy2.robjects.vectors import DataFrame as RDataFrame

from pysits.backend.functions import r_fnc_class, r_fnc_set_column
from pysits.backend.pkgs import r_pkg_arrow, r_pkg_base, r_pkg_sits


#
# Helper functions
#
def _load_arrow_table_reader_function() -> Callable[[str, list[str]], RDataFrame]:
    """Load and return an R function for reading Arrow tables with nested columns.

    This function defines and returns an R function that reads a Feather file and
    handles nested columns by unnesting them appropriately. The returned function
    takes a file path and a list of nested column names as arguments.

    Returns:
        Callable[[str, list[str]], RDataFrame]: An R function that takes a file path
            and list of nested column names as input and returns an R DataFrame with
            properly unnested columns.
    """
    rpy2_r_interface("""
        load_arrow_table <- function(path, nested_cols) {
            table <- arrow::read_feather(path)

            purrr::map_dfr(seq_len(nrow(table)), function(idx) {
                row_data <- table[idx,]

                for (col in nested_cols) {
                    row_nested <- row_data[[col]]
                    
                    # Handle arrow_list class
                    if (inherits(row_nested, "arrow_list")) {
                        row_nested <- lapply(row_nested, function(v) {
                            if (is.null(v)) return(NULL)
                            # Try to parse as JSON first
                            tryCatch({
                                parsed <- jsonlite::fromJSON(v)
                                setNames(as.character(parsed), names(parsed))
                            }, error = function(e) {
                                # If JSON parsing fails, return NULL
                                NULL
                            })
                        })
                        # If any values in row_nested are NULL, set the whole 
                        # thing to NULL
                        if (any(sapply(row_nested, is.null))) {
                            row_nested <- NULL
                        }
                    } else {
                        row_nested <- list(tidyr::unnest(
                            row_nested, 
                            cols = dplyr::everything()     
                        ))
                    }
                    
                    # Only create tibble if row_nested is not NULL
                    if (!is.null(row_nested)) {
                        row_data[[col]] <- NULL
                        row_data <- tibble::tibble(
                            row_data, 
                            !!col := row_nested
                        )
                    }
                }
                row_data
            })
        }
    """)

    return rpy2_globalenv["load_arrow_table"]


def _named_vector_to_json(x: RDataFrame, colname: str) -> RDataFrame:
    """Convert a named vector to a JSON string.

    Args:
        x (RDataFrame): R DataFrame containing a column with named vectors

        colname (str): Name of the column containing named vectors.

    Returns:
        RDataFrame: DataFrame with named vectors converted to JSON strings
    """
    # Define R code to convert named vector to JSON
    rpy2_r_interface(f"""
        named_vector_to_json <- function(x) {{
            vec_list <- lapply(x${colname}, function(v) {{
                if (is.null(names(v))) return(NULL)
                class(v) <- NULL
                json <- jsonlite::toJSON(as.list(setNames(as.character(v), names(v))), 
                                       auto_unbox=TRUE)
                class(json) <- NULL
                json
            }})
            x${colname} <- vec_list
            x
        }}
    """)

    # Call the R function and return result
    return rpy2_globalenv["named_vector_to_json"](x)


def _tibble_to_pandas_arrow(
    instance: RDataFrame,
    nested_columns: list[str] | None = None,
    table_processor: Callable[[RDataFrame], RDataFrame] | None = None,
) -> PandasDataFrame:
    """Convert an R DataFrame (tibble) to a Pandas DataFrame using Arrow format.

    This function handles the conversion of R DataFrames to Pandas DataFrames by:
    1. Creating a temporary Feather file
    2. Filtering out invalid columns (functions and NULL values)
    3. Writing valid columns to Feather format
    4. Reading back into Pandas
    5. Converting any nested columns to Pandas DataFrames

    Args:
        instance (RDataFrame): The R DataFrame (tibble) to convert.

        nested_columns (list[str] | None, optional): List of column names that
            contain nested data. Defaults to None.

    Returns:
        PandasDataFrame: The converted Pandas DataFrame.
    """
    # Create a temporary file
    tmp = tempfile.NamedTemporaryFile(suffix=".feather", delete=False)
    tmp_path = tmp.name
    tmp.close()

    # Check if instance is a empty
    if instance.nrow == 0:
        return pandas2ri.rpy2py(instance)

    # Extract columns from the data
    data_columns = r_pkg_base.colnames(instance)

    # Remove invalid columns
    data_columns_valid = []

    for data_column in data_columns:
        col = instance.rx2(data_column)

        # Remove invalid columns
        if r_fnc_class(col[0])[0] not in ["function", "NULL"]:
            data_columns_valid.append(data_column)

    # Select regular columns (using ``[]``) and convert to Pandas
    rdf_data = instance.rx(StrVector(data_columns_valid))

    # Process table
    if table_processor:
        rdf_data = table_processor(rdf_data)

    # Write to Feather format
    r_pkg_arrow.write_feather(rdf_data, tmp_path)

    # Read from Feather format
    df = feather.read_feather(tmp_path)

    # Convert nested columns to Pandas DataFrame
    if nested_columns:
        # Filter available columns
        nested_columns = [col for col in nested_columns if col in df.columns]

        # Convert nested columns to Pandas DataFrame
        for nested_column in nested_columns:
            df[nested_column] = df[nested_column].apply(
                lambda arr: PandasDataFrame.from_records(arr.tolist())
            )

    # Remove temporary file
    os.unlink(tmp_path)

    # Return value
    return df


def _pandas_to_tibble_arrow(
    instance: PandasDataFrame, nested_columns: list[str] | None = None
) -> RDataFrame:
    """Convert a Pandas DataFrame to an R DataFrame (tibble) using Arrow format.

    This function handles the conversion of Pandas DataFrames to R DataFrames by:
    1. Creating a temporary Feather file
    2. Converting nested columns to a format suitable for R
    3. Writing the data to Feather format
    4. Reading back into R using a custom Arrow table reader

    Args:
        instance (PandasDataFrame): The Pandas DataFrame to convert.

        nested_columns (list[str] | None, optional): List of column names that
            contain nested data. Defaults to None.

    Returns:
        RDataFrame: The converted R DataFrame (tibble).
    """
    instance = instance.copy(deep=True)

    tmp = tempfile.NamedTemporaryFile(suffix=".feather", delete=False)
    tmp_path = tmp.name
    tmp.close()

    # Convert nested columns to R DataFrame
    if nested_columns:
        # Filter available columns
        nested_columns = [col for col in nested_columns if col in instance.columns]

        # Convert nested columns to R DataFrame
        for nested_column in nested_columns:
            instance[nested_column] = instance[nested_column].apply(
                lambda arr: (
                    arr.to_dict(orient="list")
                    if isinstance(arr, PandasNDFrame)
                    else arr
                )
            )

    # Write to Feather
    feather.write_feather(instance, tmp_path)

    # Load Arrow table reader function
    load_arrow_table_fnc = _load_arrow_table_reader_function()

    # Read from Feather and unnest columns
    return load_arrow_table_fnc(tmp_path, nested_columns)


#
# General conversions
#
def tibble_nested_to_pandas_arrow(
    data: RDataFrame,
    nested_columns: list[str],
    table_processor: Callable[[RDataFrame], RDataFrame] | None = None,
) -> PandasDataFrame:
    """Convert any tibble to Pandas DataFrame.

    Args:
        data (rpy2.robjects.vectors.DataFrame): R (tibble/data.frame) Data frame.

    Returns:
        pandas.DataFrame: R Data Frame as Pandas.
    """
    return _tibble_to_pandas_arrow(data, nested_columns, table_processor)


def pandas_to_tibble_arrow(
    data: PandasDataFrame, nested_columns: list[str]
) -> RDataFrame:
    """Convert a pandas DataFrame to an R DataFrame object using Arrow.

    Args:
        data (pandas.DataFrame): The pandas DataFrame to convert to R.

    Returns:
        rpy2.robjects.vectors.DataFrame: The converted R DataFrame object.
    """
    return _pandas_to_tibble_arrow(data, nested_columns)


#
# SITS conversions function
#
def tibble_sits_to_pandas_arrow(data: RDataFrame) -> PandasDataFrame:
    """Convert sits tibble to Pandas DataFrame using Arrow.

    Args:
        data (rpy2.robjects.vectors.DataFrame): R (tibble/data.frame) Data frame.
    """
    # Define column order (from R)
    column_order = [
        "longitude",
        "latitude",
        "start_date",
        "end_date",
        "label",
        "cube",
        "time_series",
        "base_data",
        "predicted",
        "cluster",
        "id_sample",
        "id_neuron",
        "count",
    ]

    # Define nested columns
    nested_columns = ["time_series", "base_data", "predicted"]

    # Convert to Pandas DataFrame
    data_converted = tibble_nested_to_pandas_arrow(data, nested_columns)

    # Select columns
    columns_available = [v for v in column_order if v in data_converted.columns]

    # Return value
    return data_converted[columns_available]


def pandas_sits_to_tibble_arrow(data: PandasDataFrame) -> RDataFrame:
    """Convert sits pandas DataFrame to R DataFrame object using Arrow.

    Args:
        data (pandas.DataFrame): The pandas DataFrame to convert to R.
    """
    # Define nested columns
    nested_columns = ["time_series", "base_data", "predicted"]

    # Define data classes
    data_classes = ["sits", "tbl_df", "tbl", "data.frame"]

    if "predicted" in data.columns:
        data_classes.append("predicted")

    if "base_data" in data.columns:
        data_classes.append("sits_base")

    if "id_sample" in data.columns and "id_neuron" in data.columns:
        data_classes.append("som_clean_samples")

    # Convert to R DataFrame
    data = pandas_to_tibble_arrow(data, nested_columns)

    # Set class
    data.rclass = StrVector(data_classes)

    # Convert to R DataFrame
    return data


#
# Cube conversions function
#
def tibble_cube_to_pandas_arrow(data: RDataFrame) -> PandasDataFrame:
    """Convert sits tibble to Pandas DataFrame using Arrow.

    Args:
        data (rpy2.robjects.vectors.DataFrame): R (tibble/data.frame) Data frame.
    """
    # Define column order (from R)
    column_order = [
        "source",
        "collection",
        "satellite",
        "sensor",
        "tile",
        "xmin",
        "xmax",
        "ymin",
        "ymax",
        "crs",
        "labels",
        "file_info",
        "vector_info",
        "base_info",
    ]

    # Define nested columns
    nested_columns = ["file_info", "vector_info"]

    # Define table processor
    def table_processor(x: RDataFrame) -> RDataFrame:
        """Process table."""

        # Process ``labels`` column
        if "labels" in x.colnames:
            # Get labels column
            labels = x.rx2("labels")

            # Check if labels have names
            labels_has_names = all(
                not isinstance(label.names, NULLType) for label in labels
            )

            # If labels have names, convert to JSON
            if labels_has_names:
                x = _named_vector_to_json(x, "labels")

        return x

    # Convert to Pandas DataFrame
    data_converted = tibble_nested_to_pandas_arrow(
        data, nested_columns, table_processor
    )

    # Process base_info separately if it exists
    if "base_info" in data.colnames:
        base_info = data.rx2("base_info")
        base_info_converted = []

        # Convert each base_info item to a cube if it's not None
        for i in range(len(base_info)):
            if not isinstance(base_info[i], NULLType):
                # Convert the base_info item to a cube
                base_info_converted.append(tibble_cube_to_pandas_arrow(base_info[i]))
            else:
                base_info_converted.append(None)

        # Add converted base_info to the DataFrame
        data_converted["base_info"] = base_info_converted

    # Select columns
    columns_available = [v for v in column_order if v in data_converted.columns]

    # Return value
    return data_converted[columns_available]


def pandas_cube_to_tibble_arrow(data: PandasDataFrame) -> RDataFrame:
    """Convert sits pandas DataFrame to R DataFrame object using Arrow.

    Args:
        data (pandas.DataFrame): The pandas DataFrame to convert to R.
    """
    # Define nested columns
    nested_columns = ["labels", "file_info", "vector_info"]

    # Handle base_info separately if it exists
    base_info = None

    if "base_info" in data.columns:
        # Convert base_info to R DataFrame
        base_info = pandas_cube_to_tibble_arrow(data.base_info)

        # Drop base_info from data
        data = data.drop(columns=["base_info"])

    # Convert to R DataFrame
    data = pandas_to_tibble_arrow(data, nested_columns)

    # Add base_info back if it exists
    if base_info is not None:
        data = r_fnc_set_column(data, "base_info", base_info)

    # Set class
    data.rclass = r_pkg_sits._cube_s3class(data)

    # Return value
    return data



#
# Copyright (C) 2025 sits developers.
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <https://www.gnu.org/licenses/>.
#

"""Tmap plot module."""

import os
import tempfile

from rpy2.robjects import r as ro
from rpy2.robjects.robject import RObject

from pysits.backend.functions import r_fnc_plot
from pysits.visualization.image import show_local_image


#
# Utility function
#
def _show_tmap_plot(r_tmap_plot: RObject, image_args: dict = None, **kwargs) -> None:
    """Show an R tmap plot.

    Args:
        r_tmap_plot (rpy2.robjects.RObject): The R tmap plot object.

        **kwargs (dict): Additional keyword arguments passed to ``tmap::tmap_save``.

    Returns:
        None: Nothing.
    """
    # Create a temporary directory
    temp_dir = tempfile.mkdtemp()
    file_path = os.path.join(temp_dir, "tmap_plot.jpeg")

    # Process image args
    image_args = image_args if image_args else {}

    # Define image properties
    image_res = image_args.get("res", 300)
    image_width = int(image_args.get("width", 10) * image_res)
    image_height = int(image_args.get("height", 6) * image_res)

    kwargs.setdefault("dpi", image_res)
    kwargs.setdefault("width", image_width)
    kwargs.setdefault("height", image_height)

    # Save the tmap plot using R
    ro("tmap::tmap_save")(r_tmap_plot, filename=file_path, **kwargs)

    # Display the saved image
    show_local_image(file_path)


#
# High-level operation
#
def plot_tmap(instance: RObject, **kwargs) -> None:
    """Generates and saves a tmap plot, then displays it.

    Args:
        instance (rpy2.robjects.RObject): The R object instance for plotting.

        **kwargs (dict): Additional keyword arguments passed to ``tmap::tmap_save``.

    Returns:
        None: Nothing.
    """
    # Generate the R plot
    tmap_plot = r_fnc_plot(instance, **kwargs)

    # Save and display the plot
    _show_tmap_plot(tmap_plot)



#
# Copyright (C) 2025 sits developers.
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <https://www.gnu.org/licenses/>.
#

"""Configuration operations."""

from pysits.backend.pkgs import r_pkg_sits
from pysits.conversions.decorators import function_call
from pysits.docs import attach_doc
from pysits.models.data.base import SITStructureData


@function_call(r_pkg_sits.sits_config, SITStructureData)
@attach_doc("sits_config")
def sits_config(*args, **kwargs) -> SITStructureData:
    """Get/set sits configuration.

    ToDo:
        - Enhance result type to a Dict-like object.
    """


@function_call(r_pkg_sits.sits_config_show, lambda x: None)
@attach_doc("sits_config_show")
def sits_config_show(*args, **kwargs) -> None:
    """Show current sits configuration."""


@function_call(r_pkg_sits.sits_config_user_file, lambda x: None)
@attach_doc("sits_config_user_file")
def sits_config_user_file(*args, **kwargs) -> None:
    """Create a user configuration file."""



#
# Copyright (C) 2025 sits developers.
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <https://www.gnu.org/licenses/>.
#

"""Base visualization utilities."""

import os
import shutil
import tempfile
from typing import Any, TypeAlias

from pysits.backend.functions import r_fnc_plot
from pysits.backend.pkgs import r_pkg_grdevices
from pysits.visualization.image import show_local_image

#
# Type aliases
#
ImageArgs: TypeAlias = dict[str, int | float]


#
# Utility function
#
def _base_plot(
    data: Any,
    image_args: ImageArgs | None = None,
    multiple: bool = False,
    **kwargs: Any,
) -> None:
    """Save and show images created using base plot.

    This function creates temporary PNG files from R plot objects and displays them.
    It handles the configuration of the image device, plotting, and cleanup.
    The function handles multiple plots returned by r_fnc_plot as a ListVector.

    Args:
        data: The R object to be plotted.

        image_args: Dictionary containing image configuration parameters:
            - res (int): Resolution in DPI (default: 300)

            - width (float): Width in inches (default: 10)

            - height (float): Height in inches (default: 6)

        multiple: Whether to plot multiple plots.

        **kwargs: Additional keyword arguments passed to R's base::plot function.

    Returns:
        None: Nothing.
    """
    # Create a temporary directory
    temp_dir = tempfile.mkdtemp()

    # Process image args
    image_args = image_args if image_args else {}

    # Define image properties
    image_res = image_args.get("res", 300)
    image_width = int(image_args.get("width", 10) * image_res)
    image_height = int(image_args.get("height", 6) * image_res)

    # Plot data
    plots = r_fnc_plot(data, **kwargs)

    # Handle plots
    if multiple:
        for i, plot in enumerate(plots):
            for index, figure in enumerate(plot):
                file_path = os.path.join(temp_dir, f"base_plot_{i}_{index}.jpeg")

                # Enable image device
                r_pkg_grdevices.jpeg(
                    file=file_path,
                    width=image_width,
                    height=image_height,
                    res=image_res,
                )

                # Plot the plot object directly
                r_fnc_plot(figure)

                r_pkg_grdevices.dev_off()

                # Display saved image
                show_local_image(file_path)

    else:
        # Assuming a plot is a list of elements, we always have many elements.
        # Cases where we have a single element indicates that the plot object is
        # inside a list.
        # This is only true when the object is not a ggplot object
        while "ggplot2::ggplot" not in list(plots.rclass) and len(plots) == 1:
            plots = plots[0]

        # Create a temporary file
        file_path = os.path.join(temp_dir, "base_plot.jpeg")

        # Enable image device
        r_pkg_grdevices.jpeg(
            file=file_path, width=image_width, height=image_height, res=image_res
        )

        # Save plot using R
        r_fnc_plot(plots)

        r_pkg_grdevices.dev_off()

        # Display saved image
        show_local_image(file_path)

    # Clean up temporary directory
    shutil.rmtree(temp_dir)


#
# High-level operation
#
def plot_base(
    instance: Any,
    image_args: ImageArgs | None = None,
    multiple: bool = False,
    **kwargs: Any,
) -> None:
    """Generate and display base R plots.

    This is a high-level function that wraps ``_base_plot`` to create and display
    a plot using R's base plotting system. It handles the creation of a temporary
    PNG file and displays it.

    Args:
        instance: The R object to be plotted.

        image_args: Dictionary containing image configuration parameters:
            - res (int): Resolution in DPI (default: 300)

            - width (float): Width in inches (default: 10)

            - height (float): Height in inches (default: 6)

        multiple: Whether to plot multiple plots.

        **kwargs: Additional keyword arguments passed to R's base::plot function.

    Returns:
        None: Nothing.
    """
    # Save and display plot
    _base_plot(instance, image_args=image_args, multiple=multiple, **kwargs)



#
# Copyright (C) 2025 sits developers.
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <https://www.gnu.org/licenses/>.
#

"""Frame data models."""

from geopandas import GeoDataFrame as GeoPandasDataFrame
from pandas import DataFrame as PandasDataFrame
from rpy2.robjects.vectors import DataFrame as RDataFrame

from pysits.conversions.tibble import (
    pandas_to_tibble,
    tibble_nested_to_pandas,
    tibble_to_pandas,
)
from pysits.models.data.base import SITSData


class SITSFrameBase(SITSData):
    """Base class for SITS Data."""

    _is_updated = False
    """Whether the instance is updated."""

    def __finalize__(self, other, method=None, **kwargs):
        """Propagate metadata from another object to the current one.

        This method is called by pandas during internal operations that return a new
        object derived from an existing one, such as slicing, copying, arithmetic,
        joins, merges, and others. It ensures that any custom metadata defined in the
        ``_metadata`` attribute is preserved in the result.
        """
        if isinstance(other, SITSData):
            for name in self._metadata:
                setattr(self, name, getattr(other, name, None))
        return self

    #
    # Properties (Internal)
    #
    @property
    def _constructor(self):
        # Always return the current subclass
        return self.__class__

    def __setitem__(self, key, value):
        """Set item."""
        super().__setitem__(key, value)
        self._is_updated = True

    #
    # Convertions
    #
    def _convert_from_r(self, instance, **kwargs):
        """Convert data from R to Python."""
        return tibble_to_pandas(instance)

    #
    # Data management
    def _sync_instance(self):
        """Sync instance with R."""
        if not self._is_updated:
            return

        self._instance = pandas_to_tibble(self)

        # Update flag
        self._is_updated = False


class SITSFrame(SITSFrameBase, PandasDataFrame):
    """Base class for sits frame results."""

    #
    # Dunder methods
    #
    def __init__(self, instance, **kwargs):
        """Initializer."""
        self._instance = instance

        # Proxy instance
        if isinstance(instance, RDataFrame):
            instance = self._convert_from_r(instance)

        # Initialize super class
        PandasDataFrame.__init__(self, data=instance, **kwargs)


class SITSFrameSF(SITSFrameBase, GeoPandasDataFrame):
    """Base class for sits frame as sf."""

    #
    # Dunder methods
    #
    def __init__(self, instance, **kwargs):
        """Initializer."""
        self._instance = instance

        # Proxy instance
        if isinstance(instance, RDataFrame):
            instance = self._convert_from_r(instance)

        # Initialize super class
        GeoPandasDataFrame.__init__(self, data=instance, **kwargs)


class SITSFrameNested(SITSFrame):
    """General class for sits frame with embedded data frames."""

    # Dunder methods
    #
    def __init__(self, instance, nested_columns, **kwargs):
        """Initializer."""
        self._instance = instance

        # Proxy instance
        if isinstance(instance, RDataFrame):
            instance = self._convert_from_r(instance, nested_columns)

        # Initialize super class
        PandasDataFrame.__init__(self, data=instance, **kwargs)

    def _convert_from_r(self, instance, nested_columns, **kwargs):
        """Convert data from R to Python."""
        return tibble_nested_to_pandas(instance, nested_columns=nested_columns)
